import copy
from . import seq
from ruamel.yaml.comments import CommentedSeq
from ruamel.yaml.representer import RoundTripRepresenter

class named_list(CommentedSeq):
    """\
A class for a list of dicts, where some dicts have a 'name' item which should be unique, but others might not.
Indexing works with either number or name.
Note that updating dicts may make the list inconsistent."""
    def __init__(self,x=[]):
        CommentedSeq.__init__(self,x)

    def __getitem__(self, i):
        if isinstance(i,str):
            r = [ x for x in self if x.get('name',None)==i ]
            if len(r)>1:
                raise KeyError("There are {} elements named {}.".format(len(r),i))
            elif len(r)==0:
                raise KeyError("No element named {} found.".format(i))
            else:
                return r[0]
        else:
            return CommentedSeq.__getitem__(self,i)

    def check_consistent(self):
        """\
Checks that each name appears only once.  On failure, returns a ValueError
with (message, {failed_name: count}).  Otherwise, return with no output.
"""
        names = [ v['name'] for v in self if 'name' in v.keys() ]

        # if we have *no* names, the tests will fail, but we are obviously
        # consistent, so:
        if not names:
            return
        
        from collections import Counter
        namecounts = Counter(names)
       
        if max(namecounts.values()) > 1:
            badcounts = { n:v for n,v in namecounts.items() if v>1 }
            raise ValueError("Inconsistent named_list.", badcounts)
        
        
    def __setitem__(self, i, v):
        if isinstance(i,str):
            r = [ (ii,x) for ii,x in enumerate(self) if x.get('name',None)==i ]
            if len(r)>1:
                raise KeyError("There are {} elements named {}.".format(len(r),i))
            elif len(r)==0:
                self.append(v)
            else:
                CommentedSeq.__setitem__(self, r[0][0], v)
        else:
            CommentedSeq.__setitem__(self,i,v)

    def keys(self):
        return [ x['name'] for x in self if 'name' in x.keys() ]

    def __deepcopy__(self, memo):
        # FIXME: this is here waiting on ruamel.yaml bugfix.
        res = self.__class__()
        memo[id(self)] = res
        for k in self:
            res.append(copy.deepcopy(k))
            self.copy_attributes(res, deep=True)
        return res

RoundTripRepresenter.add_representer(named_list,
                                     RoundTripRepresenter.represent_list)

lton = { 'a': (0,),
         'b': (1, 2, 3),
         'c': (1,),
         'd': (0, 2, 3),
         'g': (2,),
         'h': (0, 1, 3),
         'k': (2, 3),
         'm': (0, 1),
         'n': (0, 1, 2, 3),
         's': (1, 2),
         't': (3,),
         'v': (0, 1, 2),
         'w': (0, 3) }
wc = {   'a': 't',
         'b': 'v',
         'c': 'g',
         'd': 'h',
         'g': 'c',
         'h': 'd',
         'k': 'm',
         'm': 'k',
         'n': 'n',
         's': 's',
         't': 'a',
         'v': 'b',
         'w': 'w' }
ntol={v:i for i,v in lton.items()}

def merge_ends(end1, end2):
    """\
Given ends end1 and end2, assuming they describe the same sticky end, merge them into a single end, combining
information from each and enforcing that the two input ends consistently make a single output end.
    """
    # Of things in the ends: fseq might need special care, everything else just needs to match.
    out = copy.deepcopy(end1)
    for i,v in end2.items():
        if i in out.keys() and i=='fseq':
            # we merge sequences    
            out[i] = seq.merge(out[i],v)
        elif i in out.keys() and out[i]!=v:
            # we insisted all others must be equal
            raise ValueError(\
                             "end1 has {}={}, end2 has {}={}".format(i,out[i],i,v))
        elif i not in out.keys():
            out[i] = copy.deepcopy(v)
    return out


def merge_endlists(endlist1, endlist2, fail_immediate=False, 
                   in_place=False ):
    """\
Given end lists `endlist1` and `endlist2`, merge the two lists, using
`merge_ends` to merge any named ends that are present in both.

Parameters
----------

endlist1: named_list of sticky ends.

endlist2: named_list OR just a list of sticky ends.  If it just a list,
    which may have multiple copies of the same named sticky end, each
    sticky end in order is merged.

fail_immediate: (default False) if True, fail immediately on a failed
merge, passing through the ValueError from merge_ends.  If False, finish
merging the two lists, then raise a ValueError listing *all* ends that
failed to merge.

in_place: (default False) if True, do merging in place in endlist1.  Note
   the merged and added ends from endlist2 will be copies regardless.

output: a merged named_list of sticky ends.


Exceptions
----------

ValueError: In the event of a failed merge, when named ends cannot be
    merged.  If fail_immediate is True, then this is passed through from
    merge_ends.  If fail_immediate is False, then the ValueError has the
    following args: 
    ("message", [exceptions], [failed_name,failed_end1,failed_end2) ...],out)
"""
    endlist1 = named_list(endlist1)

    # Check consistency of each named_list
    endlist1.check_consistent()
    try:
        endlist2.check_consistent()
    except AttributeError:
        pass
    
    if not in_place:
        out = copy.deepcopy(endlist1)
    else:
        out = endlist1
    
    exceptions = []
    errors = []
    
    for end in endlist2:
        if end['name'] in out.keys():
            try:
                out[end['name']] = merge_ends( out[end['name']], end)
            except ValueError as e:
                if fail_immediate:
                    raise e
                else:
                    exceptions.append(e)
                    errors.append((end['name'],out[end['name']],end))
        else:
            out.append(copy.deepcopy(end))
    
    if errors:
        errorstring = " ".join( [e[0] for e in errors] )
        raise ValueError(\
              "Errors merging {}".format(errorstring), exceptions, errors, out)
    
    return out

import time


class ProgressLogger(object):
    def __init__(
            self,
            logger,
            N,
            seconds_interval=60):
        self.logger = logger
        stime = time.perf_counter()
        self.stime = stime
        self.ltime = stime
        self.li = 0
        self
        self.seconds_interval = seconds_interval
        self.N = N
        self.logger.info("starting {} tasks".format(self.N))
        
    def update(self, i):
        ctime = time.perf_counter()
        if ctime - self.ltime > self.seconds_interval:
            self.logger.info(
                "finished {}/{}, {} s elapsed, {} s est remaining".format(
                    i, self.N, int(ctime - self.stime),
                    int((self.N-i)*(ctime-self.stime)/i)))
            self.ltime = ctime
            self.li = i



# End reordering code to optimize placement.

from . import anneal
from . import sensitivity as sens
import stickydesign as sd
import stickydesign.energetics as en
import numpy.random as random
import numpy as np
import random as pyrand
from copy import deepcopy
import numpy.ma as ma

def flatten(seq):
    for item in seq:
        if is_string_like(item) or not iterable(item):
            yield item
        else:
            for subitem in flatten(item):
                yield subitem

def ecomp(x):
    if x[-1]=='/':
        return x[:-1]
    else:
        return x+'/'

class FseqState:
    def __init__(self, seqs=None):
        if not seqs:
            self.seqs = {}
        else:
            self.seqs = seqs
    def copy(self):
        return FseqState({'DT': self.seqs['DT'].copy(), 'TD': self.seqs['TD'].copy()})

class cachedarray:
    def __init__(self, func, shape):
        self.arr = ma.masked_all(shape)
        self.func = func
    def __getitem__(self, index):
        if self.arr.mask[index]:
            self.arr[index] = self.func(*index)
        return self.arr[index]

from copy import deepcopy
class FastState:
    def __init__(self, state):
        self.state = state
    def copy(self):
        return deepcopy(self)
    def __getitem__(self, it):
        return self.state[it]


class EndSystemFseq:
    def __init__(self, tilesys, newends=None, pairs=None, energetics=None):
        # Set up variables, etc.
        if not energetics:
            self.ef = en.energetics_santalucia(mismatchtype='max')
        else:
            self.ef = energetics
        tilesys = deepcopy(tilesys)
        self.ends = tilesys['ends']
        self.tiles = tilesys['tiles']
        self.tilesystem = tilesys
        
        if not pairs:
            pairs = sens.consolidate_pairs( sens.senspairs(tilesys), comcomp=1, onlytop=True )

        self.names = {}
        fseqsTD, self.names['TD'] = (list(x) for x in zip(*[ [end['fseq'].lower(),end['name']] for end in self.ends if end['type'] == 'TD']))
        fseqsDT, self.names['DT'] = (list(x) for x in zip(*[ [end['fseq'].lower(),end['name']] for end in self.ends if end['type'] == 'DT']))
        self.seqs = {}
        self.seqs['TD'] = sd.endarray(fseqsTD,'TD')
        self.seqs['DT'] = sd.endarray(fseqsDT,'DT')
        self.initstate = FastState({'DT': np.arange(0,len(self.seqs['DT'])), 'TD': np.arange(0,len(self.seqs['TD']))})
        self.enlocs = {}
        for i, endn in enumerate(self.names['TD']):
            self.enlocs[endn] = (i,'TD')
        for i, endn in enumerate(self.names['DT']):
            self.enlocs[endn] = (i,'DT')        

        # ends that can be reordered.
        if newends:
            self.mutableTD = [ i for i,t in [self.enlocs[x] for x in newends] if t=='TD']
            self.mutableDT = [ i for i,t in [self.enlocs[x] for x in newends] if t=='DT']
        else:
            self.mutableTD = range(0,len(fseqsTD))
            self.mutableDT = range(0,len(fseqsDT))
            
        # Get the mean non-spurious interaction
        self.meangse = 0.5*( np.mean(self.ef.matching_uniform( self.seqs['TD'] ))+np.mean(self.ef.matching_uniform( self.seqs['DT'] )) )
        self.mult = {'1NGO': np.exp(-2.0*self.meangse), '2NGO': np.exp(-1.65*self.meangse), '1GO': np.exp(-1.5*self.meangse), '2GO': np.exp(-1.1*self.meangse)}
        
        self.pairdict = {}
        for pairclass,memberset in pairs.items():
            for x,y in memberset:
                self.pairdict[(x,ecomp(y))] = pairclass
                self.pairdict[(y,ecomp(x))] = pairclass
        tdsh = ( len(self.seqs['TD']), len(self.seqs['TD']) )
        dtsh = ( len(self.seqs['DT']), len(self.seqs['DT']) )
        self.ecache_cc = {  'TD': cachedarray( lambda x,y: self.ef.uniform(self.seqs['TD'][x:x+1].comps,self.seqs['TD'][y:y+1].comps) , tdsh ),
                            'DT': cachedarray( lambda x,y: self.ef.uniform(self.seqs['DT'][x:x+1].comps,self.seqs['DT'][y:y+1].comps) , dtsh ) }
        self.ecache_ce = {  'TD': cachedarray( lambda x,y: self.ef.uniform(self.seqs['TD'][x:x+1].comps,self.seqs['TD'][y:y+1].ends) , tdsh ),
                            'DT': cachedarray( lambda x,y: self.ef.uniform(self.seqs['DT'][x:x+1].comps,self.seqs['DT'][y:y+1].ends) , dtsh ) }
        self.ecache_ec = {  'TD': cachedarray( lambda x,y: self.ef.uniform(self.seqs['TD'][x:x+1].ends,self.seqs['TD'][y:y+1].comps) , tdsh ),
                            'DT': cachedarray( lambda x,y: self.ef.uniform(self.seqs['DT'][x:x+1].ends,self.seqs['DT'][y:y+1].comps) , dtsh ) }
        self.ecache_ee = {  'TD': cachedarray( lambda x,y: self.ef.uniform(self.seqs['TD'][x:x+1].ends,self.seqs['TD'][y:y+1].ends) , tdsh ),
                            'DT': cachedarray( lambda x,y: self.ef.uniform(self.seqs['DT'][x:x+1].ends,self.seqs['DT'][y:y+1].ends) , dtsh ) }
    def slowseqs(self, state):
        "Give the state as the slow version would have"
        return {'DT': self.seqs['DT'][state['DT']], 'TD': self.seqs['TD'][state['TD']] }

    def mutate(self, state):
        # Start by deciding to swap TD or DT ends.
        if random.rand() > 1.0*len(self.mutableTD)/(len(self.mutableDT)+
                                                    len(self.mutableTD)):
            a = random.choice(self.mutableDT)
            b = random.choice(self.mutableDT)
            state['DT'][[a,b]] = state['DT'][[b,a]]
        else:
            a = random.choice(self.mutableTD)
            b = random.choice(self.mutableTD)
            state['TD'][[a,b]] = state['TD'][[b,a]]

    def score(self, state):
        
        sc = 0.0
        
        for (xn,yn),pairclass in self.pairdict.items():
            
            # set comp flags
            xc = False
            yc = False
            if xn[-1] == '/':
                xc = True
                xn = xn[:-1]
            if yn[-1] == '/':
                yc = True
                yn = yn[:-1]
                
            # get end indexes and types
            xi,xt = self.enlocs[xn]
            yi,yt = self.enlocs[yn]
            #print "%s, %s: (%s %s)" % (xn,yn,xt,yt)
            # skip if not same type
            if xt != yt: continue
            
            if yc and xc:
                val = self.ecache_cc[xt][state[xt][xi],state[yt][yi]] # self.ef.uniform(state.seqs[xt][xi:xi+1].comps,state.seqs[yt][yi:yi+1].comps)[0]
            elif xc:
                val = self.ecache_ce[xt][state[xt][xi],state[yt][yi]] # self.ef.uniform(state.seqs[xt][xi:xi+1].comps,state.seqs[yt][yi:yi+1].ends)[0]
            elif yc:
                val = self.ecache_ec[xt][state[xt][xi],state[yt][yi]] # state[f.uniform(state.seqs[xt][xi:xi+1].ends,state.seqs[yt][yi:yi+1].comps)[0]
            else:
                val = self.ecache_ee[xt][state[xt][xi],state[yt][yi]] # self.ef.uniform(state.seqs[xt][xi:xi+1].ends,state.seqs[yt][yi:yi+1].ends)[0]
            
            sc += self.mult[pairclass]*np.exp( val )
    
        return sc   

            
wcd = {   'a': 't',
         'b': 'v',
         'c': 'g',
         'd': 'h',
         'g': 'c',
         'h': 'd',
         'k': 'm',
         'm': 'k',
         'n': 'n',
         's': 's',
         't': 'a',
         'v': 'b',
         'w': 'w' }
         
def wc(seqstr):
    return ''.join(wcd[x] for x in reversed(seqstr))



import ruamel.yaml as yaml
from ruamel.yaml.comments import CommentedMap
from ruamel.yaml.representer import RoundTripRepresenter
from collections import Counter 
from .util import named_list, merge_endlists
from . import tiletypes
from functools import reduce
import operator
import copy


class TileSet(CommentedMap):
    def __init__(self, val={}):
        CommentedMap.__init__(self, val)
        if 'ends' in self.keys():
            self['ends'] = named_list(self['ends'])
        else:
            self['ends'] = named_list()
        if 'tiles' in self.keys():
            self['tiles'] = named_list(self['tiles'])
        else:
            self['tiles'] = named_list()

    def check_consistent(self):
        # * END LIST The end list itself must be consistent.
        # ** Each end must be of understood type
        # ** Each end must have a valid sequence or no sequence
        # ** There must be no more than one instance of each name
        # ** WARN if there are ends with no namecounts
        # * TILE LIST
        # ** each tile must be of understood type (must parse)
        for tile in self['tiles']:
            parsed = tiletypes.tfactory.parse(tile)
            # ** the tile type edotparen must be consistent, if it has one
            if parsed.edotparen:
                tiletypes.check_edotparen_consistency(parsed.edotparen)
            else:
                log.warning("tile type {} has no edotparen".format(tile['type']))
            # ** each tile must have no sequence, or a valid sequence
            if 'fullseqs' in tile.keys():
                parsed.check_sequence()
            # ** each tile must have the right number of ends
            if 'ends' in tile.keys():
                assert len(parsed._endtypes) == len(tile['ends'])
        # ** ends in the tile list must be consistent (must merge)
        endsfromtiles = tiletypes.endlist_from_tilelist(self['tiles'])
        # ** there must be no more than one tile with each name
        self['tiles'].check_consistent()
        # ** WARN if any end that appears does not have a complement used or vice versa
        # ** WARN if there are tiles with no name
        # * TILE + END
        # ** The tile and end lists must merge validly
        # (checks sequences, adjacents, types, complements)
        merge_endlists(self['ends'], endsfromtiles)
        
        # ** WARN if tilelist has end references not in ends
        # ** WARN if merge is not equal to the endlist
        # ** WARN if endlist has ends not used in tilelist
        # * ADAPTERS / SEEDS
        # ** seeds must be of understood type
        # ** adapter locations must be valid
        # ** each adapter must have no sequence or a consistent sequence
        # *** the RH strand must match the associated tile
        # *** the ends in the sequence must match the ends in the endlist
        # *** the LH sequence must be validly binding to both RH and origami
        # ** each adapter must have valid definition, which means for us:
        # *** if both tile mimic and ends are specified, they must match
    
    def summary(self):
        self.check_consistent()
        info = {'ntiles': len(self['tiles']),
                'nends':  len(self['ends']),
                'ntends': len(tiletypes.endlist_from_tilelist(self['tiles'])),
                'tns':    " ".join(x['name'] for x in self['tiles'] if 'name' in x.keys()),
                'ens':    " ".join(x['name'] for x in self['ends'] if 'name' in x.keys()),
                'name':   " {}".format(self['info']['name']) if \
                                       ('info' in self.keys() and \
                                        'name' in self['info'].keys()) else ""}
        tun = sum( 1 for x in self['tiles'] if 'name' not in x.keys() )
        if tun > 0:
            info['tns'] += " ({} unnamed)".format(tun)
        eun = sum( 1 for x in self['ends'] if 'name' not in x.keys() )
        if eun > 0:
            info['ens'] += " ({} unnamed)".format(eun)
        return "TileSet{name}: {ntiles} tiles, {nends} ends, {ntends} ends in tiles.\nTiles: {tns}\nEnds:  {ens}".format(**info)

    def __str__( self ):
        return self.summary()

    def copy(self):
        return copy.deepcopy(self)

    def __deepcopy__(self, memo):
        # type: (Any) -> Any
        res = self.__class__()
        memo[id(self)] = res
        for k in self:
            res[k] = copy.deepcopy(self[k])
            self.copy_attributes(res, deep=True)
        return res

    def dump(self, stream):
        return yaml.round_trip_dump(self, stream)
    
RoundTripRepresenter.add_representer(TileSet,
                                     RoundTripRepresenter.represent_dict)


def load_tileset_dict(*args, **kwargs):
    return TileSet(yaml.round_trip_load(*args, **kwargs))



#!/usr/bin/env python

# Python module for simulated annealing - anneal.py - v1.0 - 2 Sep 2009
# 
# Copyright (c) 2009, Richard J. Wagner <wagnerr@umich.edu>
# 
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
import logging as log
"""
This module performs simulated annealing to find a state of a system that
minimizes its energy.

An example program demonstrates simulated annealing with a traveling
salesman problem to find the shortest route to visit the twenty largest
cities in the United States.
"""

# How to optimize a system with simulated annealing:
# 
# 1) Define a format for describing the state of the system.
# 
# 2) Define a function to calculate the energy of a state.
# 
# 3) Define a function to make a random change to a state.
# 
# 4) Choose a maximum temperature, minimum temperature, and number of steps.
# 
# 5) Set the annealer to work with your state and functions.
# 
# 6) Study the variation in energy with temperature and duration to find a
# productive annealing schedule.
# 
# Or,
# 
# 4) Run the automatic annealer which will attempt to choose reasonable values
# for maximum and minimum temperatures and then anneal for the allotted time.

import copy, math, random, sys, time
import numpy as np

def round_figures(x, n):
	"""Returns x rounded to n significant figures."""
	return round(x, int(n - math.ceil(math.log10(abs(x)))))

def time_string(seconds):
	"""Returns time in seconds as a string formatted HHHH:MM:SS."""
	s = int(round(seconds))  # round to nearest second
	h, s = divmod(s, 3600)   # get hours and remainder
	m, s = divmod(s, 60)     # split remainder into minutes and seconds
	return '%4i:%02i:%02i' % (h, m, s)

class Annealer:
	"""Performs simulated annealing by calling functions to calculate
	energy and make moves on a state.  The temperature schedule for
	annealing may be provided manually or estimated automatically.
	"""
	def __init__(self, energy, move):
		self.energy = energy  # function to calculate energy of a state
		self.move = move      # function to make a random change to a state
	
	def anneal(self, state, Tmax, Tmin, steps, updates=0):
		"""Minimizes the energy of a system by simulated annealing.
		
		Keyword arguments:
		state -- an initial arrangement of the system
		Tmax -- maximum temperature (in units of energy)
		Tmin -- minimum temperature (must be greater than zero)
		steps -- the number of steps requested
		updates -- the number of updates to print during annealing
		
		Returns the best state and energy found."""
		
		step = 0
		start = time.time()
		
		def update(T, E, acceptance, improvement):
			"""Prints the current temperature, energy, acceptance rate,
			improvement rate, elapsed time, and remaining time.
			
			The acceptance rate indicates the percentage of moves since the last
			update that were accepted by the Metropolis algorithm.  It includes
			moves that decreased the energy, moves that left the energy
			unchanged, and moves that increased the energy yet were reached by
			thermal excitation.
			
			The improvement rate indicates the percentage of moves since the
			last update that strictly decreased the energy.  At high
			temperatures it will include both moves that improved the overall
			state and moves that simply undid previously accepted moves that
			increased the energy by thermal excititation.  At low temperatures
			it will tend toward zero as the moves that can decrease the energy
			are exhausted and moves that would increase the energy are no longer
			thermally accessible."""
			
			elapsed = time.time() - start
			if step == 0:
				log.info( ' Temperature        Energy    Accept   Improve     Elapsed   Remaining  Step' )
				log.info( '%12.2e  %12.2e                      %s            ' % \
					(T, E, time_string(elapsed) ) )
			else:
				remain = ( steps - step ) * ( elapsed / step )
				log.info( '%12.2e  %12.2e  %7.2f%%  %7.2f%%  %s  %s   %s' % \
					(T, E, 100.0*acceptance, 100.0*improvement,
						time_string(elapsed), time_string(remain), step) )
		
		# Precompute factor for exponential cooling from Tmax to Tmin
		if Tmin <= 0.0:
			log.info( 'Exponential cooling requires a minimum temperature greater than zero.' )
			sys.exit()
		Tfactor = -math.log( float(Tmax) / Tmin )
		
		# Note initial state
		T = Tmax
		E = self.energy(state)
		prevState = state.copy()
		prevEnergy = E
		bestState = state.copy()
		bestEnergy = E
		trials, accepts, improves = 0, 0, 0
		if updates > 0:
			updateWavelength = updates
			update(T, E, None, None)
		
		# Attempt moves to new states
		while step < steps:
			step += 1
			T = Tmax * math.exp( Tfactor * step / steps )
			self.move(state)
			E = self.energy(state)
			dE = E - prevEnergy
			trials += 1
			if dE > 0.0 and math.exp(-dE/T) < random.random():
				# Restore previous state
				state = prevState.copy()
				E = prevEnergy
			else:
				# Accept new state and compare to best state
				accepts += 1
				if dE < 0.0:
					improves += 1
				prevState = state.copy()
				prevEnergy = E
				if E < bestEnergy:
					bestState = state.copy()
					bestEnergy = E
			if updates > 1:
				if step // updateWavelength > (step-1) // updateWavelength:
					update(T, E, float(accepts)/trials, float(improves)/trials)
					trials, accepts, improves = 0, 0, 0
		
		# Return best state and energy
		return bestState, bestEnergy
	
	def auto(self, state, minutes, steps=2000):
		"""Minimizes the energy of a system by simulated annealing with
		automatic selection of the temperature schedule.
		
		Keyword arguments:
		state -- an initial arrangement of the system
		minutes -- time to spend annealing (after exploring temperatures)
		steps -- number of steps to spend on each stage of exploration
		
		Returns the best state and energy found."""
		
		def run(state, T, steps):
			"""Anneals a system at constant temperature and returns the state,
			energy, rate of acceptance, and rate of improvement."""
			E = self.energy(state)
			prevState = state.copy()
			prevEnergy = E
			accepts, improves = 0, 0
			for step in range(steps):
				self.move(state)
				E = self.energy(state)
				dE = E - prevEnergy
				if dE > 0.0 and math.exp(-dE/T) < random.random():
					state = prevState.copy()
					E = prevEnergy
				else:
					accepts += 1
					if dE < 0.0:
						improves += 1
					prevState = state.copy()
					prevEnergy = E
			return state, E, float(accepts)/steps, float(improves)/steps
		
		step = 0
		start = time.time()
		
		log.info( 'Attempting automatic simulated anneal...' )
		
		# Find an initial guess for temperature
		T = 0.0
		E = self.energy(state)
		while T == 0.0:
			step += 1
			self.move(state)
			T = abs( self.energy(state) - E )
		
		log.info( 'Exploring temperature landscape:' )
		log.info( ' Temperature        Energy    Accept   Improve     Elapsed' )
		def update(T, E, acceptance, improvement):
			"""Prints the current temperature, energy, acceptance rate,
			improvement rate, and elapsed time."""
			elapsed = time.time() - start
			log.info( '%12.2e  %12.2e  %7.2f%%  %7.2f%%  %s' % (T, E, 100.0*acceptance, 100.0*improvement, time_string(elapsed)))
		
		# Search for Tmax - a temperature that gives 98% acceptance
		state, E, acceptance, improvement = run(state, T, steps)
		step += steps
		while acceptance > 0.98:
			T = round_figures(T/1.5, 2)
			state, E, acceptance, improvement = run(state, T, steps)
			step += steps
			update(T, E, acceptance, improvement)
		while acceptance < 0.98:
			T = round_figures(T*1.5, 2)
			state, E, acceptance, improvement = run(state, T, steps)
			step += steps
			update(T, E, acceptance, improvement)
		Tmax = T
		
		# Search for Tmin - a temperature that gives 0% improvement
		while improvement > 0.0:
			T = round_figures(T/1.5, 2)
			state, E, acceptance, improvement = run(state, T, steps)
			step += steps
			update(T, E, acceptance, improvement)
		Tmin = T
		
		# Calculate anneal duration
		elapsed = time.time() - start
		duration = round_figures(int(60.0 * minutes * step / elapsed), 2)
		
		# Perform anneal
		log.info( 'Annealing from %.2f to %.2f over %i steps:' % (Tmax, Tmin, duration) )
		return self.anneal(state, Tmax, Tmin, duration, 20)

if __name__ == '__main__':
	"""Test annealer with a traveling salesman problem."""
	
	# List latitude and longitude (degrees) for the twenty largest U.S. cities
	cities = { 'New York City': (40.72,74.00), 'Los Angeles': (34.05,118.25),
	'Chicago': (41.88,87.63), 'Houston': (29.77,95.38),
	'Phoenix': (33.45,112.07), 'Philadelphia': (39.95,75.17),
	'San Antonio': (29.53,98.47), 'Dallas': (32.78,96.80),
	'San Diego': (32.78,117.15), 'San Jose': (37.30,121.87),
	'Detroit': (42.33,83.05), 'San Francisco': (37.78,122.42),
	'Jacksonville': (30.32,81.70), 'Indianapolis': (39.78,86.15),
	'Austin': (30.27,97.77), 'Columbus': (39.98,82.98),
	'Fort Worth': (32.75,97.33), 'Charlotte': (35.23,80.85),
	'Memphis': (35.12,89.97), 'Baltimore': (39.28,76.62) }
	
	def distance(a, b):
		"""Calculates distance between two latitude-longitude coordinates."""
		R = 3963  # radius of Earth (miles)
		lat1, lon1 = math.radians(a[0]), math.radians(a[1])
		lat2, lon2 = math.radians(b[0]), math.radians(b[1])
		return math.acos( math.sin(lat1)*math.sin(lat2) +
			math.cos(lat1)*math.cos(lat2)*math.cos(lon1-lon2) ) * R
	
	def route_move(state):
		"""Swaps two cities in the route."""
		a = random.randint( 0, len(state)-1 )
		b = random.randint( 0, len(state)-1 )
		state[a], state[b] = state[b], state[a]
	
	def route_energy(state):
		"""Calculates the length of the route."""
		e = 0
		for i in range(len(state)):
			e += distance( cities[state[i-1]], cities[state[i]] )
		return e
	
	# Start with the cities listed in random order
	state = cities.keys()
	random.shuffle(state)
	
	# Minimize the distance to be traveled by simulated annealing with a
	# manually chosen temperature schedule
	annealer = Annealer(route_energy, route_move)
	state, e = annealer.anneal(state, 10000000, 0.01, 18000*len(state), 9)
	while state[0] != 'New York City':
		state = state[1:] + state[:1]  # rotate NYC to start
	log.info( "%i mile route:" % route_energy(state) )
	for city in state:
		log.info( "\t", city )
	
	# Minimize the distance to be traveled by simulated annealing with an
	# automatically chosen temperature schedule
	state, e = annealer.auto(state, 4)
	while state[0] != 'New York City':
		state = state[1:] + state[:1]  # rotate NYC to start
	log.info( "%i mile route:" % route_energy(state) )
	for city in state:
		log.info( "\t", city )
	
	sys.exit()



from .tiletypes import gettile, check_edotparen_sequence, check_edotparen_consistency
import warnings
import copy

class seed_base:
    def create_adapter_sequence_diagram( self, adapterdef ):
        # At the moment, there are only three potential structures, so I'm just
        # going to put them here.
        ttype = 'tile_adapter_5up'
        if 'extra' in adapterdef.keys():
            ttype += '_'+adapterdef['extra']
        
        from lxml import etree
        import pkg_resources
        import os
        base_svg = etree.parse( pkg_resources.resource_stream(__name__, os.path.join('seqdiagrambases',ttype+'.svg') ) )

        try:
            s = adapterdef['seqs'][1]
            l = adapterdef['seqs'][0]
        except KeyError:
            s = adapterdef['fullseqs'][1]
            l = adapterdef['fullseqs'][0]
        name = adapterdef['name']
        if 'extra' not in adapterdef.keys():
            strings = [(s[0:5]+'--'+s[5:13])[::-1],s[13:21]+'--'+s[21:26], \
                      (l[0:8]+'--'+l[8:16])[::-1],l[16:24][::-1],l[24:32],l[32:40]+'--'+l[40:48], \
                      adapterdef['ends'][0],adapterdef['ends'][1],name]
        elif adapterdef['extra'] == '1h':
            hp = s[0:13]; s = s[13:]
            strings = [hp[0:5]+'-'+hp[5:7]+'-'+hp[7:9], \
                      (hp[9:11]+'-'+hp[11:13]+'-'+s[0:5]+'--'+s[5:13])[::-1],s[13:21]+'--'+s[21:26], \
                      (l[0:8]+'--'+l[8:16])[::-1],l[16:24][::-1],l[24:32],l[32:40]+'--'+l[40:48], \
                      adapterdef['ends'][1],name]
        elif adapterdef['extra'] == '2h':
            strings = [(s[0:5]+'--'+s[5:13])[::-1],s[13:21]+'--'+s[21:26]+'-'+s[26:28]+'-'+s[28:30], \
                      (s[30:32]+'-'+s[32:34]+'-'+s[34:39])[::-1], \
                      (l[0:8]+'--'+l[8:16])[::-1],l[16:24][::-1],l[24:32],l[32:40]+'--'+l[40:48], \
                      adapterdef['ends'][0],name]

        texts = base_svg.findall( "//{http://www.w3.org/2000/svg}text" )

        for text, string in zip(texts,strings):
            text.text = string

        return base_svg.xpath('/svg:svg/svg:g',namespaces={'svg':'http://www.w3.org/2000/svg'})[0]

class seed_tileadapts(seed_base):
    needspepper = False
    def create_adapter_sequences( self, tileset ):
        tset = copy.deepcopy( tileset )

        for adapter in tset['seed']['adapters']:
            tile_to_mimic = self.getmimic( tset, adapter )
            if tile_to_mimic['type'] == 'tile_daoe_5up':
                # Tile is single. If we have ends, check that they match.
                if ( 'ends' in adapter.keys() ) and ( adapter['ends'] != tile_to_mimic['ends'][1:3] ):
                    raise ValueError("adapter {} and tile base {} ends don't match: adapter has {}, tile has {}".format(
                        adapter['name'], tile_to_mimic['name'], adapter['ends'], tile_to_mimic['ends'][1:3] ))
                adapter_strand_short = tile_to_mimic['fullseqs'][-1]
                tile_long_strand = tile_to_mimic['fullseqs'][-2]
            elif tile_to_mimic['type'] == 'tile_daoe_doublehoriz_35up' \
                    or tile_to_mimic['type'] == 'tile_daoe_doublevert_35up':
                if ( 'ends' in adapter.keys() ) and ( adapter['ends'] != tile_to_mimic['ends'][2:4] ):
                    raise ValueError("adapter {} and tile base {} ends don't match: adapter has {}, tile has {}".format(
                        adapter['name'], tile_to_mimic['name'], adapter['ends'], tile_to_mimic['ends'][2:4] ))
                adapter_strand_short = tile_to_mimic['fullseqs'][-1]
                tile_long_strand = tile_to_mimic['fullseqs'][-2]

            adapter['seqs'] = [ tile_long_strand[0:8]+self.cores[adapter['loc']-1]+tile_long_strand[40:48], 
                                adapter_strand_short ]

        return tset

    def getmimic( self, tileset, adapter ):
        try:
            mimic = gettile( tileset, adapter['tilebase'] )
        except KeyError:
            warnings.warn("tilestrand format is deprecated; strand number will be ignored")
            mimic = gettile( tileset, adapter['tilestrand'][0] )
        return mimic

    def getmimic_full( self, tileset, adapter ):
        import re
        mimic = self.getmimic( tileset, adapter )
        out = {}
        e1 = None
        e2 = None
        if mimic['type'] == 'tile_daoe_5up':
            out['tilestrand'] = mimic['fullseqs'][3]
            out['tileends'] = mimic['ends'][1:3]
            if 'extra' in mimic.keys():
                e1 = re.search(r'2([A-Za-z]+)',mimic['extra'])
                e2 = re.search(r'3([A-Za-z]+)',mimic['extra'])
        elif mimic['type'] == 'tile_daoe_doublehoriz_35up' or \
           mimic['type'] == 'tile_daoe_doublevert_35up':
            out['tilestrand'] = mimic['fullseqs'][5]
            out['tileends'] = mimic['ends'][2:4]
            if 'extra' in mimic.keys():
                e1 = re.search(r'3([A-Za-z]+)',mimic['extra'])
                e2 = re.search(r'4([A-Za-z]+)',mimic['extra'])
        else:
            raise ValueError("Tile type {} not known for adapter.".format(
                mimic['type']), adapter, mimic)
        if e1:
            if e1[1] == 'h':
                out['extra']='1h'
            else:
                raise ValueError("Tile has unknown extra {}".format(e1),
                                 adapter, mimic)
        if e2:
            if e2[1] == 'h':
                out['extra']='2h'
            else:
                raise ValueError("Tile has unknown extra {}".format(e2),
                                 adapter, mimic)
        out['tile'] = mimic
        return out
            
                
    
    def check_consistent( self, tileset ):
        import re
        for adapt in tileset['seed']['adapters']:
            mimic = self.getmimic_full( tileset, adapt )
            assert mimic.get('extra') == adapt.get('extra')
            assert mimic['tileends'] == adapt['ends']
            

    edotparen_adapt = "8(32.8(+5.16)5."
    edotparen_adapt_2h = "8(32.8(+5.16)7(4.7)"
    edotparen_adapt_1h = "8(32.8(+7(4.7)16)5."
    
    def check_sequence( self, tileset ):
        import re
        check_edotparen_consistency( self.edotparen_adapt )
        check_edotparen_consistency( self.edotparen_adapt_1h )
        check_edotparen_consistency( self.edotparen_adapt_2h )
        self.check_consistent( tileset )
        for adapt in tileset['seed']['adapters']:
            mimic = self.getmimic_full(tileset, adapt)
            seqs = "+".join(adapt['seqs'])
            if 'extra' not in adapt.keys():
                check_edotparen_sequence(
                    self.edotparen_adapt, seqs )
            elif adapt['extra'] == '1h':
                check_edotparen_sequence(
                    self.edotparen_adapt_1h, seqs )
            elif adapt['extra'] == '2h':
                check_edotparen_sequence(
                    self.edotparen_adapt_2h, seqs )
            else:
                raise ValueError("Unknown extra", adapt['extra'])
            assert adapt['seqs'][0][8:(8+32)] == self.cores[adapt['loc']-1]
                
        
class tallrect_tileadapts(seed_tileadapts):
    cores = ['CAGGAACGGTACGCCATTAAAGGGATTTTAGA',
             'CTACATTTTGACGCTCACGCTCATGGAAATAC',
             'CCAGCAGAAGATAAAAAATACCGAACGAACCA',
             'GCCGTCAATAGATAATCAACTAATAGATTAGA',
             'ACTTCTGAATAATGGATGATTGTTTGGATTAT',
             'GAAGATGATGAAACAAAATTACCTGAGCAAAA',
             'CATAGGTCTGAGAGACGTGAATTTATCAAAAT',
             'GAAAAAGCCTGTTTAGGGAATCATAATTACTA',
             'ACGCGCCTGTTTATCAGTTCAGCTAATGCAGA',
             'GCTTATCCGGTATTCTAAATCAGATATAGAAG',
             'AACGTCAAAAATGAAAAAACGATTTTTTGTTT',
             'GCAGATAGCCGAACAATTTTTAAGAAAAGTAA',
             'AGACAAAAGGGCGACAGGTTTACCAGCGCCAA',
             'GCGTCAGACTGTAGCGATCAAGTTTGCCTTTA',
             'GTCAGACGATTGGCCTCAGGAGGTTGAGGCAG',
             'TGAAAGTATTAAGAGGCTATTATTCTGAAACA']

class triangle_side2(seed_tileadapts):
    # Note: this list is currently reversed because
    # the cadnano file for the triangle has 5' ends of
    # the last staple above 3' ends, while our adapter
    # scheme has 3' ends above 5' ends. 
    cores = list(reversed( \
            ['AGAGAGTACCTTTAATCCAACAGGTCAGGATT',
             'TAAGAGGAAGCCCGAAATTGCATCAAAAAGAT',
             'CCCCCTCAAATGCTTTTAAATATTCATTGAAT',
             'TACCAGACGACGATAATATCATAACCCTCGTT',
             'CGTTAATAAAACGAACTGGGAAGAAAAATCTA',
             'TAACAAAGCTGCTCATATTACCCAAATCAACG',
             'ATTGTGTCGAAATCCGTATCATCGCCTGATAA',
             'CGAGGGTAGCAACGGCAAAGACAGCATCGGAA',
             'TCTCCAAAAAAAAGGCTTTTTCACGTTGAAAA',
             'TTTCGTCACCAGTACAGTACCGTAACACTGAG',
             'TCAGTACCAGGCGGATATTAGCGGGGTTTTGC',
             'GATACAGGAGTGTACTATACATGGCTTTTGAT',
             'ACCAGAGCCGCCGCCACGCCACCAGAACCACC',
             'GCGTTTGCCATCTTTTCATAGCCCCCTTATTA',
             'ATGAAACCATCGATAGGCCGGAAACGTCACCA',
             'ATCACCGTCACCGACTTCATTAAAGGTGAATT']))

class tallrect_endadapts(seed_base):
    needspepper = True
    cores = tallrect_tileadapts.cores
    def create_pepper_input_files( self, seeddef, importlist, compstring ):
        for adapter in seeddef['adapters']:
            adaptertype = 'tile_adapter_5up'
            e = [[],[]]
            e[0].append('origamicore_{0}'.format(adapter['loc']))
            for end in adapter['ends']:
                if (end == 'hp'):
                    continue # skip hairpins, etc that aren't designed by stickydesign
                e[0].append( 'e_'+end.replace('/','*') )
                if end[-1]=='/':
                    a = 'c_'+end[:-1]+'*'
                else:
                    a = 'a_'+end
                e[1].append( a )
            s1 = " + ".join(e[0])
            s2 = " + ".join(e[1])
            if 'extra' in adapter.keys():
                adaptertype+='_'+adapter['extra']
            compstring += "component {} = {}: {} -> {}\n".format(adapter['name'],adaptertype,s1,s2)
            importlist.add( adaptertype )
        return (importlist, compstring)

seedtypes = { 
        'tallrect_tileadapts': tallrect_tileadapts(),
        'tallrect_endadapts': tallrect_endadapts(),
        'triangle_side2': triangle_side2()
        }
seedtypes['longrect'] = seedtypes['tallrect_tileadapts']
seedtypes['longrect_old'] = seedtypes['tallrect_endadapts']



import re
import numpy as np
import string
from io import StringIO
import pandas as pd

def loadflake(xgrowstring,onlytiles=False):
    tiles = np.genfromtxt(
        StringIO(                                       # genfromtxt needs an io object
            re.sub(r'(\[|\])','',                       # remove [ and ]
                re.sub(r'; \.\.\.','',xgrowstring))),   # remove ; ... at line ends
        skip_header=4,                                  # remove first lines
        skip_footer=1,                                  # remove end junk
        dtype='uint'
        )
    if onlytiles:
        return tiles
    data = pd.Series(
        np.genfromtxt(StringIO(xgrowstring.split('\n')[2]))[1:-1],
        index = ['gmc','gse','k','time','tiles','mismatches','events','perimeter','g','dgbonds']
        )
    return {'data': data, 'tiles': tiles}



# Utility functions for sequences
import string

_L_TO_N = { 'a': frozenset((0,)),
            'b': frozenset((1, 2, 3)),
            'c': frozenset((1,)),
            'd': frozenset((0, 2, 3)),
            'g': frozenset((2,)),
            'h': frozenset((0, 1, 3)),
            'k': frozenset((2, 3)),
            'm': frozenset((0, 1)),
            'n': frozenset((0, 1, 2, 3)),
            'r': frozenset((0,2)),
            's': frozenset((1, 2)),
            't': frozenset((3,)),
            'v': frozenset((0, 1, 2)),
            'w': frozenset((0, 3)),
            'y': frozenset((1,3))}
# FIXME: X (synonym for N) is not allowed.

_N_TO_L = { v: i for i,v in _L_TO_N.items() }

_AMBBASES = frozenset( _L_TO_N.keys() - {'a','c','g','t'} )

def is_null(seq):
    """Return True if a sequence consists only of Ns, or is empty. 
    Return False otherwise."""
    check_bases(seq)
    return set(seq.lower()).issubset(set('n').union(set(string.whitespace)))

def is_definite(seq):
    """Return True if a sequence consists only of defined bases.  Return
    False otherwise.  If blank, return False.
    """
    check_bases(seq)
    if set(seq.lower()).issubset(set(string.whitespace)):
        return False
    return set(seq.lower()).issubset({'a','g','c','t'}.union(set(string.whitespace))) 

def check_bases(seq):
    if not set(seq.lower()).issubset(set(_L_TO_N.keys()).union(set(string.whitespace))):
        raise ValueError("Sequence has unknown bases.")

def count_ambiguous(seq):
    """Return the number of ambiguous bases in a sequence."""
    check_bases(seq)
    return sum( 1 for x in seq.lower() if x in _AMBBASES )

def length(seq):
    """Return the length of a sequence, stripping whitespace.  This does not handle
    extended labels, etc."""
    check_bases(seq)
    return len( seq.translate( str.maketrans('','',string.whitespace) ) )


def merge(seq1, seq2):
    """Merge two sequences together, returning a single sequence that
    represents the constraint of both sequences.  If the sequences
    can't be merged, raise a MergeConflictError.

    FIXME: this needs to intelligently handle case and whitespace.
    """
    
    check_bases(seq1)
    check_bases(seq2)
    if len(seq1) != len(seq2):
        raise MergeConflictError(seq1,seq2,'length',len(seq1),len(seq2))
    
    out = []
    for i,(n1,n2) in enumerate(zip(seq1.lower(),seq2.lower())):
        try:
            out.append(_N_TO_L[frozenset.intersection(_L_TO_N[n1], _L_TO_N[n2])])
        except KeyError as e:
            if e.args[0] == frozenset():
                raise MergeConflictError(seq1,seq2,i,n1,n2) from None
            else:
                raise e
    return ''.join(out)


class MergeConflictError(ValueError):
    """
    Merge of items failed because of conflicting information.
    Arguments are (item1, item2, location or property, value1, value2)
    """
    
class MergeConflictsError(ValueError):
    """
    Merge of multiple items failed because individual merges
    raised MergeConflictErrors.
    Arguments are ([list of MergeConflictErrors])
    """



import unittest
import alhambra.seq as seq
import random
import string

class SeqTests(unittest.TestCase):
    seq1 = "aggagtacc"
    seq2 = "aggn"
    seq3 = "gngc"
    seq4 = "hggc"
    defbases = frozenset(('a','g','t','c'))
    ambbases = frozenset(seq._L_TO_N.keys()) - defbases
    allbases = frozenset(seq._L_TO_N.keys()) 
    def test_merge_correct(self):
        # same sequence
        self.assertEqual(seq.merge(self.seq1,self.seq1),self.seq1)
        # constraints on both sides
        self.assertEqual(seq.merge(self.seq2,self.seq4),'aggc')
        self.assertEqual(seq.merge(self.seq4,self.seq2),'aggc')
    def test_merge_conflicts(self):
        with self.assertRaises(seq.MergeConflictError):
            seq.merge(self.seq1,self.seq2)
        with self.assertRaises(seq.MergeConflictError):
            seq.merge(self.seq2,self.seq3)
        with self.assertRaises(seq.MergeConflictError):
            seq.merge(self.seq3,self.seq4)
    def test_is_null(self):
        self.assertTrue( seq.is_null("") )
        self.assertTrue( seq.is_null("nnnnnnnnnnnnnnnnNNNNNNN") )
        self.assertFalse( seq.is_null("nnagaagannna") )
        self.assertFalse( seq.is_null("nnnnnnnynnnn") )
    def test_is_definite(self):
        for _ in range(0,100):
            self.assertTrue(
                seq.is_definite(
                    ''.join(random.choices(tuple(self.defbases),k=50))) )
            self.assertFalse(
                seq.is_definite(
                    ''.join(random.choices(tuple(self.ambbases),k=50))) )
            self.assertFalse(
                seq.is_definite(
                    ''.join(
                        random.choices(
                            tuple(frozenset.union(self.defbases,self.ambbases)),k=50))+'b'))
            # FIXME: this last test needs that extra b to ensure that random doesn't choose only
            # definite bases.  Test coverage is hindered.
    def test_space(self):
        self.assertTrue( seq.is_null('Nn nn  nn\tnn\nnn') )
        self.assertTrue( seq.is_null('       \t\n  \n') )
        self.assertFalse( seq.is_definite('    \t\n  \n'))
        self.assertFalse( seq.is_null('nnnnnnnn      \t a \n nnnnnn'))
        self.assertFalse( seq.is_definite('ag  n n\t\nn') )
        self.assertTrue( seq.is_definite('ag  c t\t\ntt') )
    def test_length(self):
        n = random.randint(0,100)
        s = list(
            ''.join(random.choices(tuple(self.allbases),
                                   k=n)) + \
            ''.join(random.choices(tuple(string.whitespace),
                                   k=random.randint(0,100))))
        random.shuffle(s)
        self.assertEqual( seq.length(''.join(s)), n )
        self.assertEqual( seq.length('\n\t'), 0 )
        self.assertEqual( seq.length(''), 0 )
    def test_count_ambiguous(self):
        for _ in range(0,100):
            n = random.randint(0,1000)
            s = list(
                ''.join(random.choices(tuple(self.ambbases),
                                       k=n)) + \
                ''.join(random.choices(tuple(self.defbases),
                                       k=random.randint(0,1000))) + \
                ''.join(random.choices(tuple(string.whitespace),
                                       k=random.randint(0,1000))))
            random.shuffle(s)
            self.assertEqual( seq.count_ambiguous(''.join(s)), n )
        # FIXME: deal with uppercase.
if __name__ == '__main__':
    unittest.main()



# End reordering code to optimize placement.

from . import anneal
from . import sensitivity as sens
import stickydesign as sd
import stickydesign.energetics as en
import numpy.random as random
import numpy as np
import random as pyrand
from copy import deepcopy

def flatten(seq):
    for item in seq:
        if is_string_like(item) or not iterable(item):
            yield item
        else:
            for subitem in flatten(item):
                yield subitem

def ecomp(x):
    if x[-1]=='/':
        return x[:-1]
    else:
        return x+'/'

class FseqState:
    def __init__(self, seqs=None):
        if not seqs:
            self.seqs = {}
        else:
            self.seqs = seqs
    def copy(self):
        return FseqState({'DT': self.seqs['DT'].copy(), 'TD': self.seqs['TD'].copy()})

class EndSystemFseq:
    def __init__(self, tilesys, pairs=None, energetics=None):
        # Set up variables, etc.
        if not energetics:
            self.ef = en.energetics_santalucia(mismatchtype='max')
        else:
            self.ef = energetics
        tilesys = deepcopy(tilesys)
        self.ends = tilesys['ends']
        self.tiles = tilesys['tiles']
        self.tilesystem = tilesys
        
        if not pairs:
            pairs = sens.consolidate_pairs( sens.senspairs(tilesys), comcomp=1, onlytop=True )

        self.initstate = FseqState()
        self.names = {}
        fseqsTD, self.names['TD'] = (list(x) for x in zip(*[ [end['fseq'].lower(),end['name']] for end in self.ends if end['type'] == 'TD']))
        fseqsDT, self.names['DT'] = (list(x) for x in zip(*[ [end['fseq'].lower(),end['name']] for end in self.ends if end['type'] == 'DT']))
        self.initstate.seqs['TD'] = sd.endarray(fseqsTD,'TD')
        self.initstate.seqs['DT'] = sd.endarray(fseqsDT,'DT')
        self.enlocs = {}
        for i, endn in enumerate(self.names['TD']):
            self.enlocs[endn] = (i,'TD')
        for i, endn in enumerate(self.names['DT']):
            self.enlocs[endn] = (i,'DT')        
        
        # Get the mean non-spurious interaction
        self.meangse = 0.5*( np.mean(self.ef.matching_uniform( self.initstate.seqs['TD'] ))+np.mean(self.ef.matching_uniform( self.initstate.seqs['DT'] )) )
        self.mult = {'1NGO': np.exp(-2.0*self.meangse), '2NGO': np.exp(-1.65*self.meangse), '1GO': np.exp(-1.5*self.meangse), '2GO': np.exp(-1.1*self.meangse)}
        
        self.pairdict = {}
        for pairclass,memberset in pairs.items():
            for x,y in memberset:
                self.pairdict[(x,ecomp(y))] = pairclass
                self.pairdict[(y,ecomp(x))] = pairclass
        
        
    def mutate(self, state):
        # Start by deciding to swap TD or DT ends.
        if random.rand() > 1.0*len(state.seqs['TD'])/(len(state.seqs['DT'])+len(state.seqs['TD'])):
            t = 'DT'
        else:
            t = 'TD'
        
        en = len(state.seqs[t])
        
        a = random.randint( 0, en )
        b = random.randint( 0, en )
        state.seqs[t][[a,b],:] = state.seqs[t][[b,a],:]

    def score(self, state):
        
        sc = 0.0
        
        for (xn,yn),pairclass in self.pairdict.items():
            
            # set comp flags
            xc = False
            yc = False
            if xn[-1] == '/':
                xc = True
                xn = xn[:-1]
            if yn[-1] == '/':
                yc = True
                yn = yn[:-1]
                
            # get end indexes and types
            xi,xt = self.enlocs[xn]
            yi,yt = self.enlocs[yn]
            #print "%s, %s: (%s %s)" % (xn,yn,xt,yt)
            # skip if not same type
            if xt != yt: continue
            
            if yc and xc:
                val = self.ef.uniform(state.seqs[xt][xi:xi+1].comps,state.seqs[yt][yi:yi+1].comps)[0]
            elif xc:
                val = self.ef.uniform(state.seqs[xt][xi:xi+1].comps,state.seqs[yt][yi:yi+1].ends)[0]
            elif yc:
                val = self.ef.uniform(state.seqs[xt][xi:xi+1].ends,state.seqs[yt][yi:yi+1].comps)[0]
            else:
                val = self.ef.uniform(state.seqs[xt][xi:xi+1].ends,state.seqs[yt][yi:yi+1].ends)[0]
            
            sc += self.mult[pairclass]*np.exp( val )
    
        return sc   

            
wcd = {   'a': 't',
         'b': 'v',
         'c': 'g',
         'd': 'h',
         'g': 'c',
         'h': 'd',
         'k': 'm',
         'm': 'k',
         'n': 'n',
         's': 's',
         't': 'a',
         'v': 'b',
         'w': 'w' }
         
def wc(seqstr):
    return ''.join(wcd[x] for x in reversed(seqstr))



from . import stickyends
from . import tiletypes
import stickydesign.plots as sdplots
import stickydesign as sd
from collections import Counter
from stickydesign import EnergeticsDAOE
from matplotlib import pylab
import numpy as np

DEFAULT_REGION_ENERGETICS = EnergeticsDAOE(temperature=33,
                                           coaxparams=False,
                                           danglecorr=False)


def plot_se_hists(tileset,
                  all_energetics=None,
                  energetics_names=None,
                  title=None, **kwargs):

    if all_energetics is None:
        all_energetics = stickyends.DEFAULT_MULTIMODEL_ENERGETICS

    if energetics_names is None:
        energetics_names = stickyends.DEFAULT_MM_ENERGETICS_NAMES

    if 'ends' in tileset.keys():
        ends = tileset['ends']
    else:
        ends = tileset
        
    if title is None:
        # FIXME
        title = 'Title'

    td = sd.endarray([x['fseq'] for x in ends
                      if x['type'] == 'TD'], 'TD')

    dt = sd.endarray([x['fseq'] for x in ends
                      if x['type'] == 'DT'], 'DT')

    sdplots.hist_multi([td, dt],
                       all_energetics,
                       energetics_names,
                       title, **kwargs)


def plot_adjacent_regions(tileset,
                          energetics=None):

    if energetics is None:
        energetics = DEFAULT_REGION_ENERGETICS

    regions = [tiletypes.tfactory.parse(t)._side_bound_regions
               for t in tileset['tiles']]
    regions = [[x.lower() for x in y] for y in regions]
    allregions = sum(regions, [])
    count = [[Counter(x) for x in y] for y in regions]
    gc_count = [[x['g']+x['c'] for x in c] for c in count]
    gc_counts = sum(gc_count, [])

    ens = energetics.matching_uniform(sd.endarray(allregions, 'DT'))

    pylab.figure(figsize=(10, 4))
    pylab.subplot(121)
    pylab.hist(gc_counts,
               bins=np.arange(min(gc_counts)-0.5, max(gc_counts)+0.5))
    pylab.title('G/C pairs in arms')
    pylab.ylabel('# of 15 nt arms')
    pylab.xlabel('# of G/C pairs')
    pylab.subplot(122)
    pylab.hist(ens)
    pylab.title('ΔG, T=33, no coaxparams/danglecorr')
    pylab.ylabel('# of 8 nt regions')
    pylab.xlabel('stickydesign ΔG')
    pylab.suptitle('8 nt end-adjacent region strengths')


def plot_side_strands(tileset,
                      energetics=None):

    if energetics is None:
        energetics = DEFAULT_REGION_ENERGETICS

    regions = [tiletypes.tfactory.parse(t)._short_bound_full
               for t in tileset['tiles']]
    regions = [[x.lower() for x in y] for y in regions]
    allregions = sum(regions, [])
    count = [[Counter(x) for x in y] for y in regions]
    gc_count = [[x['g']+x['c'] for x in c] for c in count]
    gc_counts = sum(gc_count, [])

    ens = energetics.matching_uniform(sd.endarray(allregions, 'DT'))

    pylab.figure(figsize=(10, 4))
    pylab.subplot(121)
    pylab.hist(gc_counts,
               bins=np.arange(min(gc_counts)-0.5, max(gc_counts)+0.5))
    pylab.title('G/C pairs in arms')
    pylab.ylabel('# of 15 nt arms')
    pylab.xlabel('# of G/C pairs')
    pylab.subplot(122)
    pylab.hist(ens)
    pylab.title('ΔG, T=33, no coaxparams/danglecorr')
    pylab.ylabel('# of 16 nt regions')
    pylab.xlabel('stickydesign ΔG')
    pylab.suptitle('16 nt arm region strengths')



import stickydesign as sd
import numpy as np
import re
import copy
from . import stickyends


def generate_xgrow_dict(ts,
                        perfect=False,
                        rotate=False,
                        energetics=None):
    
    # Combine ends and tile-specified adjacents
    newtiles = []
    newends = []
    doubleends = []
    doubles = []
    vdoubleends = []
    vdoubles = []
    ts = copy.deepcopy(ts)
    newtiles.append({'name': 'origami',
                     'edges': ['origami', 'origami', 'origami', 'origami'],
                     'stoic': 0,
                     'color': 'white'})

    atiles = [None]*16
    for tilename in ts['seed']['use_adapters']:
        try:
            tile = [ x for x in ts['seed']['adapters'] if x['name']==tilename ][0]
        except IndexError:
            raise Exception("Can't find {}".format(tilename))
        newtile = {}
        newtile['edges'] = [ 'origami' ] +  [ re.sub('/','_c',x) for x in tile['ends'] ] + [ 'origami' ]
        newtile['name'] = tile['name']
        newtile['stoic'] = 0
        newtile['color'] = 'white'
        atiles[tile['loc']-1] = newtile
    for tile in atiles:
        if tile:
            newtiles.append(tile)
        else:
            newtiles.append( { 'name': 'emptyadapt', 'edges': ['origami',0,0,'origami'], 'stoic': 0, 'color': 'white'} )


    if rotate:
        rotatedtiles = []
        for tile in ts['tiles']:
            if tile['type'] == 'tile_daoe_3up' or tile['type'] == 'tile_daoe_5up':
                newtile = copy.deepcopy(tile)
                newtile['name']+='_lrf'
                newtile['ends']=[tile['ends'][x] for x in (1,0,3,2)]
                rotatedtiles.append(newtile)
                newtile = copy.deepcopy(tile)
                newtile['name']+='_udf'
                newtile['type']='tile_daoe_'+{'5up':'3up','3up':'5up'}[tile['type'][-3:]]
                newtile['ends']=[tile['ends'][x] for x in (3,2,1,0)]
                rotatedtiles.append(newtile)
                newtile = copy.deepcopy(tile)
                newtile['name']+='_bf'
                newtile['type']='tile_daoe_'+{'5up':'3up','3up':'5up'}[tile['type'][-3:]]
                newtile['ends']=[tile['ends'][x] for x in (2,3,0,1)]
                rotatedtiles.append(newtile)
            elif tile['type'] == 'tile_daoe_doublehoriz_35up':
                newtile = copy.deepcopy(tile)
                newtile['name']+='_lrf'
                newtile['type']='tile_daoe_doublevert_53up'
                newtile['ends']=[tile['ends'][x] for x in (2,1,0,5,4,3)]
                rotatedtiles.append(newtile)
                newtile = copy.deepcopy(tile)
                newtile['name']+='_udf'
                newtile['type']='tile_daoe_doublevert_53up'
                newtile['ends']=[tile['ends'][x] for x in (5,4,3,2,1,0)]
                rotatedtiles.append(newtile)
                newtile = copy.deepcopy(tile)
                newtile['name']+='_bf'
                newtile['ends']=[tile['ends'][x] for x in (3,4,5,0,1,2)]
                rotatedtiles.append(newtile)
            elif tile['type'] == 'tile_daoe_doublevert_35up':
                newtile = copy.deepcopy(tile)
                newtile['name']+='_lrf'
                newtile['type']='tile_daoe_doublehoriz_53up'
                newtile['ends']=[tile['ends'][x] for x in (2,1,0,5,4,3)]
                rotatedtiles.append(newtile)
                newtile = copy.deepcopy(tile)
                newtile['name']+='_udf'
                newtile['type']='tile_daoe_doublehoriz_53up'
                newtile['ends']=[tile['ends'][x] for x in (5,4,3,2,1,0)]
                rotatedtiles.append(newtile)
                newtile = copy.deepcopy(tile)
                newtile['name']+='_bf'
                newtile['ends']=[tile['ends'][x] for x in (3,4,5,0,1,2)]
                rotatedtiles.append(newtile)

        ts['tiles'] += rotatedtiles
    
    for tile in ts['tiles']:
        if tile['type'] == 'tile_daoe_3up' or tile['type'] == 'tile_daoe_5up':
            newtile = {}
            newtile['edges'] = [ re.sub('/','_c',x) for x in tile['ends'] ]
            if 'name' in tile: newtile['name'] = tile['name']
            if 'conc' in tile: newtile['stoic'] = tile['conc']
            if 'color' in tile: newtile['color'] = tile['color']
            newtiles.append(newtile)

        if tile['type'] == 'tile_daoe_doublehoriz_35up' or tile['type'] == 'tile_daoe_doublehoriz_53up':
            newtile1 = {}
            newtile2 = {}
            newtile1['edges'] = [ re.sub('/','_c',x) for x in tile['ends'][0:1] ] \
                + [ tile['name']+'_db' ] \
                + [ re.sub('/','_c',x) for x in tile['ends'][4:] ]
            newtile2['edges'] = [ re.sub('/','_c',x) for x in tile['ends'][1:4] ] \
                + [ tile['name']+'_db' ]            
            newtile1['name'] = tile['name']+'_left'
            newtile2['name'] = tile['name']+'_right'
                        
            doubleends.append( tile['name']+'_db' )
            doubles.append( (newtile1['name'], newtile2['name']) )
            
            if 'conc' in tile: 
                newtile1['stoic'] = tile['conc']
                newtile2['stoic'] = tile['conc']
                
            if 'color' in tile: 
                newtile1['color'] = tile['color']
                newtile2['color'] = tile['color']
                
            newtiles.append(newtile1)
            newtiles.append(newtile2)
        if tile['type'] == 'tile_daoe_doublevert_35up' or tile['type'] == 'tile_daoe_doublevert_53up':
            newtile1 = {}
            newtile2 = {}
            newtile1['edges'] = [ re.sub('/','_c',x) for x in tile['ends'][0:2] ] \
                + [ tile['name']+'_db' ] \
                + [ re.sub('/','_c',x) for x in tile['ends'][5:] ]
            newtile2['edges'] = [ tile['name']+'_db' ] + [ re.sub('/','_c',x) for x in tile['ends'][2:5] ] 
            newtile1['name'] = tile['name']+'_top'
            newtile2['name'] = tile['name']+'_bottom'
                        
            vdoubleends.append( tile['name']+'_db' )
            vdoubles.append( (newtile1['name'], newtile2['name']) )
            
            if 'conc' in tile: 
                newtile1['stoic'] = tile['conc']
                newtile2['stoic'] = tile['conc']
                
            if 'color' in tile: 
                newtile1['color'] = tile['color']
                newtile2['color'] = tile['color']
                
            newtiles.append(newtile1)
            newtiles.append(newtile2)
    
    newends.append( { 'name': 'origami', 'strength': 100 } )
 
    for end in doubleends:
        newends.append( { 'name': end, 'strength': 10 } )
    for end in vdoubleends:
        newends.append( { 'name': end, 'strength': 10 } )

    gluelist = []
    if not perfect: 
        glueends = {'DT': [], 'TD': []}
        for end in ts['ends']:
            newends.append( { 'name': end['name'], 'strength': 0 } )
            newends.append( { 'name': end['name']+'_c', 'strength': 0 } )
            if (end['type'] == 'TD') or (end['type'] == 'DT'):
                glueends[end['type']].append((end['name'],end['fseq']))
        
        if energetics:
            ef = energetics
        else:
            ef = stickyends.DEFAULT_ENERGETICS
       
        eavg = {}
        for t in ['DT', 'TD']:
            names, fseqs = zip(*glueends[t])
            ea = sd.endarray(fseqs, t)
            eavg[t] = np.average( ef.matching_uniform( ea ) )
        eavg_combined = ( eavg['DT'] + eavg['TD'] ) / 2.0
        
        for t in ['DT','TD']:
            names, fseqs = zip(*glueends[t])
            allnames = names + tuple( x+'_c' for x in names )
            ea = sd.endarray(fseqs, t)
            ar = sd.energy_array_uniform(ea,ef) / eavg_combined
            for i1,n1 in enumerate(names):
                for i2,n2 in enumerate(allnames):
                    gluelist.append([n1,n2,max(float(ar[i1,i2]),0.0)])

         
    else:
        if 'ends' not in ts.keys():
            ts['ends']=[]
        endsinlist = set( e['name'] for e in ts['ends'] )
        endsintiles = set()
        for tile in ts['tiles']:
            endsintiles.update( re.sub('/','',e) for e in tile['ends'] if e != 'hp')
        for end in ts['ends'] + list({'name': e} for e in endsintiles):
            newends.append( { 'name': end['name'], 'strength': 0 } )
            newends.append( { 'name': end['name']+'_c', 'strength': 0 } )
            gluelist.append([end['name'],end['name']+'_c',1.0]) 
            

    
    newends.append( {'name': 'hp', 'strength': 0} )

    xga = {}
    xga['doubletiles'] = [ list(x) for x in doubles ]
    xga['vdoubletiles'] = [ list(x) for x in vdoubles ]
    xga.update( ts['xgrow_options'] )
    xga.update( ts['xgrow_options'] )
    #if not perfect:
    #    xga['gse_calc_avg'] = eavg_combined
     
        
    sts = { 'tiles': newtiles, 'bonds': newends, 'xgrowargs': xga, 'glues': gluelist }
    
    return sts



# Structured Xgrow libary for reading, writing and converting Structured Xgrow (STXG) files.
# If you're just manipulating stxg files, you probably don't need this (just use yaml), unless
# you want to use the dump function, which produces somewhat cosmetically better output. The 
# library is primarily useful for converting to and from xgrow tile files.
#
# If you just want to run xgrow with stxg files, consider using xgrow-wrap.

version = '0.0.1'

import ruamel.yaml as yaml
import re
from io import StringIO
import datetime
import copy
#import warning

# Option names and their argument types.
keyvaloptions = [ ('block', int), ('size', int), ('rand', str), ('k', float), ('Gmc', float), ('Gse', float),
                  ('Gas', float),('Gam', float),('Gae', float),('Gfc', float),('T', float),('blast_rate_alpha', float),
                  ('blast_rate_beta', float),('blast_rate_gamma', float),('seed', str), ('update_rate', int),
                  ('tracefile', str), ('untiltiles', str), ('tmax', float),('emax', int),('smax', int),
                  ('smin', int), ('untiltilescount', str),('clean_cycles', int),('error_radius', float),
                  ('datafile', str),('arrayfile', str),('exportfile', str),('importfile', str),('min_strength', float) ]
keyopts = [x[0] for x in keyvaloptions]

# Options that need special handling.
specialoptions = ['stoic', 'nowindow', 'doubletile', 'vdoubletile', 'fission']

# Options that take true/fale values.
truefalseoptions = ['pause','wander','periodic']

# Classes for controlling flow/block style of various yaml elements. These are used in dump()
class blockseq( dict ): pass
def blockseq_rep(dumper, data):
    return dumper.represent_mapping( u'tag:yaml.org,2002:seq', data, flow_style=False )
class flowmap( dict ): pass
def flowmap_rep(dumper, data):
    return dumper.represent_mapping( u'tag:yaml.org,2002:map', data, flow_style=True )
yaml.add_representer(blockseq, blockseq_rep)
yaml.add_representer(flowmap, flowmap_rep)


def dump( tileset, *xargs, **pargs ):
    """
    stxg.dump( tileset, ... ) is a wrapper for yaml.dump that slightly prettifies
    the output taking into account what tilesets usually look like. It passes
    all arguments onto yaml.dump.
    """
    
    # Make tile and bond sections flow-style
    tileset['tiles'] = [ flowmap(x) for x in tileset['tiles'] ]
    tileset['bonds'] = [ flowmap(x) for x in tileset['bonds'] ]
    
    # If xgrowargs is there, make it block-style
    if 'xgrowargs' in tileset.keys(): 
        tileset['xgrowargs'] = blockseq(tileset['xgrowargs'])
    
    return yaml.dump(tileset, *xargs, **pargs)


def load( stream, *xargs, **pargs ):
    """
    stxg.load( stream, ... ) is a wrapper for yaml.load that currently does
    nothing else but call yaml.load
    """
    return yaml.load( stream, *xargs, **pargs )


def to_xgrow( stxg, stream=None ):
    """
    Given an stxg structure/dict, create an xgrow tile file, and return it
    as a string. Alternatively, if the stream= parameter is given a stream,
    write the file to it instead.
    """
    # If there's no stream to write to, create a StringIO to write to:
    if stream:
        xgrowf = stream
    else:
        xgrowf = StringIO()
    
    # Add some comments:
        
    for n,tile in enumerate(stxg['tiles']):
        tileedges = [str(x) for x in tile['edges']]
        for e in tileedges:
            if e != '0' and e not in [x['name'] for x in stxg['bonds']]:
                stxg['bonds'].append( { 'name': e, 'strength': 1 } )
    # Start writing the tileset
    try:
        xgrowf.write("num tile types=%d\n" % len(stxg['tiles']))
        xgrowf.write("num binding types=%d\n" % len(stxg['bonds']))
    except KeyError as e:
        raise ConvertError("Missing section {0}.".format(e.message))
  
    bondhasname = [ 'name' in x.keys() for x in stxg['bonds'] ]
    
    if all(bondhasname):
        xgrowf.write("binding type names={ %s }\n" % " ".join([x['name'] for x in stxg['bonds']]))
    elif any(bondhasname):
        raise ConvertError("Some bonds have names, but not all. All bonds must be either named or unnamed.")
    
    xgrowf.write("tile edges={\n")
    
    for n,tile in enumerate(stxg['tiles']):
        xgrowf.write("{ %s }" % " ".join([str(x) for x in tile['edges']]))
        if 'stoic' in tile:
            xgrowf.write("[%g]" % tile['stoic'])
        if 'color' in tile:
            xgrowf.write("(%s)" % tile['color'])
        xgrowf.write("   % ")
        if 'name' in tile:
            xgrowf.write(tile['name'])
        xgrowf.write(" (tile #%d)\n" % n)
    
    xgrowf.write("}\n")
    
    xgrowf.write("binding strengths={ %s }\n" % " ".join([str(x['strength']) for x in stxg['bonds']]))
    
    if 'glues' in stxg:
        for x1,x2,g in [tuple(x) for x in stxg['glues']]:
            # dereference
            if type(x1) != int:
                try:
                    x1 = [n for n,z in enumerate(stxg['bonds']) if z['name']==x1][0]+1
                except IndexError:
                    print(x2)
                    raise ValueError(x2)
            if type(x2) != int:
                try:
                    x2 = [n for n,z in enumerate(stxg['bonds']) if z['name']==x2][0]+1
                except IndexError:
                    print(x2)
                    raise ValueError(x2)
            xgrowf.write("g(%d,%d)=%g\n" % (x1,x2,float(g)))
    
    if 'xgrowargs' in stxg.keys():
        for key,val in stxg['xgrowargs'].items():
            if key in keyopts:
                xgrowf.write("%s=%s\n" % (key, str(val)))
            elif key in truefalseoptions:
                if val:
                    xgrowf.write(key+'\n')
            elif key == 'window':
                if not val:
                    xgrowf.write('-nw\n')
            elif key == 'doubletiles':
                for x1,x2 in val:
                    # dereference
                    if type(x1) != int:
                        try:
                            x1 = [n for n,z in enumerate(stxg['tiles']) if z['name']==x1][0]+1
                        except IndexError:
                            print(x1)
                            raise ValueError(x1)
                    if type(x2) != int:
                        try:
                            x2 = [n for n,z in enumerate(stxg['tiles']) if z['name']==x2][0]+1
                        except IndexError:
                            print(x2)
                            raise ValueError(x2)           
                    xgrowf.write( "doubletile=%d,%d\n" % (x1,x2) )
            elif key == 'vdoubletiles':
                for x1,x2 in val:
                    # dereference
                    if type(x1) != int:
                        try:
                            x1 = [n for n,z in enumerate(stxg['tiles']) if z['name']==x1][0]+1
                        except IndexError:
                            print(x1)
                            raise ValueError(x1)
                    if type(x2) != int:
                        try:
                            x2 = [n for n,z in enumerate(stxg['tiles']) if z['name']==x2][0]+1
                        except IndexError:
                            print(x2)
                            raise ValueError(x2)           
                    xgrowf.write( "vdoubletile=%d,%d\n" % (x1,x2) )
            elif key == 'fission':
                if val == 'off':
                    xgrowf.write("no_fission\n")
                elif val == 'on':
                    xgrowf.write("fission\n")
                elif val == 'chunk':
                    xgrowf.write("chunk_fission\n")
    xgrowf.write("% Tileset created by stxg.py version {0} on {1}\n".format(version, str(datetime.date.today())) )
    
    if stream:
        return None
    else:
        return xgrowf.getvalue()
        
        
def from_xgrow( xgst ):
    """
    Given a string of an xgrow tile file, convert it to stxg.
    
    Note that the conversion may require some work afterward.
    """
    
    ntre = re.compile(r"^\s*num tile types=(\d+)", re.MULTILINE)
    nbre = re.compile(r"^\s*num binding types=(\d+)", re.MULTILINE)
    bsre = re.compile(r"^\s*binding strengths\s*=\s*{([^}]+)}", re.MULTILINE)
    tere = re.compile(r"^\s*tile edges\s*=[\s\n]*{((?:[^{}]*{[^{}]+}[^{}]*)+)}", re.MULTILINE)
    bnre = re.compile(r"^\s*binding type names\s*=\s*{([^}]+)}", re.MULTILINE)
    glre = re.compile(r"^\s*g\(\s*(\d+)\s*,\s*(\d+)\s*\)\s*=\s*(\S+)",re.MULTILINE)
    
    xgs = re.sub( r"%[^\n]\n", "\n", xgst )
    
    m = ntre.search(xgs)
    if not m:
        #warning.warn("Expected number of tile types.")
        pass
    ntiles = int( m.group(1) )
        
    m = nbre.search(xgs)
    if not m:
        #warning.warn("Expected number of binding types.")
        pass
    nbind = int( m.group(1) )
    
    bsstring = bsre.search(xgs).group(1)
    tstring = tere.search(xgs).group(1)
    bnmatch = bnre.search(xgs)
    if bnmatch:
        bnstring = bnmatch.group(1)
        bnames = bnstring.split()
    
    bstrengths = bsstring.split()
    tilevals = re.findall(r"{\s*(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s*}(?:\[(\S+)\])?(?:\((\S+)\))?",tstring)
    
    tiles = []
    for tv in tilevals:
        tile = {}
        tile['edges'] = []
        for e in tv[0:4]:
            try:
                e = int(e)
            except:
                pass
            tile['edges'].append(e)
        if tv[4]!='':
            tile['stoic'] = float(tv[4])
        if tv[5]!='':
            tile['color'] = tv[5]
        tiles.append(tile)
        
    if bnmatch:
        assert len(bnames) == len(bstrengths)
        bonds = [ {'name': name, 'strength': float(strength)} for name, strength in zip(bnames,bstrengths) ]
    else:
        bonds = [ {'strength': float(strength)} for strength in bstrengths ]
    
    assert len(bonds) == nbind
    assert len(tiles) == ntiles
    
    ts = {'tiles': tiles, 'bonds': bonds}
    
    glues = {}
    m = glre.findall(xgs)
    if m:
        for x1,x2,s in m:
            if bnmatch:
                glues[(bnames[int(x1)-1], bnames[int(x2)-1])] = float(s)
            else:
                glues[int(x1),int(x2)] = float(s)
        ts['glues'] = glues
    
    xgrowargs = {}
    for opt,opttype in keyvaloptions:
        m = re.findall( r"^\s*"+opt+r"\s*=\s*(\S+)", xgs, re.MULTILINE )
        if len(m)>0:
            xgrowargs[opt] = opttype(m[-1])
            
    # fission
    m = re.findall( r"^\s*(chunk_fission|fission|no_fission)", xgs, re.MULTILINE )
    if len(m)>0:
        if m[-1] == 'chunk_fission':
            xgrowargs['fission'] = 'chunk'
            
        elif m[-1] == 'fission':
            xgrowargs['fission'] = 'on'
        else:
            xgrowargs['fission'] = 'off'
    
    # doubletile
    m = re.findall( r"^\s*doubletile\s*=\s*(\d+)\s*,\s*(\d+)", xgs, re.MULTILINE )
    if m:
        xgrowargs['doubletiles'] = [ [int(x1),int(x2)] for x1,x2 in m ]
    
    if re.search(r"^\s*-nw", xgs, re.MULTILINE ):
        xgrowargs['window'] = False
        
    ts['xgrowargs'] = xgrowargs
    return ts  
    
#

# Stuff originally from yamltostxg.py. This is even less polished than everything above.

def from_yaml_endadj( ts, perfect=False, rotate=False ):
    import stickydesign as sd
    import stickydesign.energetics as en
    import numpy as np
    
    # Combine ends and tile-specified adjacents
    newtiles = []
    newends = []
    endslist = set()
    doubleends = []
    doubles = []
    vdoubleends = []
    vdoubles = []
    
    newtiles.append( { 'name': 'origami', 'edges': ['origami','origami','origami','origami'], 'stoic': 0, 'color': 'white'} )
    
    atiles = [None]*16
    for tilename in ts['seed']['use_adapters']:
        tile = [ x for x in ts['seed']['adapters'] if x['name']==tilename ][0]
        newtile = {}
        newtile['edges'] = [ 'origami' ] +  [ re.sub('/','_c',x) for x in tile['ends'] ] + [ 'origami' ]
        newtile['name'] = tile['name']
        newtile['stoic'] = 0
        newtile['color'] = 'white'
        atiles[tile['loc']-1] = newtile
    for tile in atiles:
        if tile:
            newtiles.append(newtile)
        else:
            newtiles.append( { 'name': 'emptyadapt', 'edges': ['origami',0,0,'origami'], 'stoic': 0, 'color': 'white'} )

    if rotate:
        rotatedtiles = []
        for tile in ts['tiles']:
            if tile['type'] == 'tile_daoe_3up' or tile['type'] == 'tile_daoe_5up':
                newtile = copy.copy(tile)
                newtile['name']+='_lrf'
                newtile['ends']=[tile['ends'][x] for x in (1,0,3,2)]
                rotatedtiles.append(newtile)
                newtile = copy.copy(tile)
                newtile['name']+='_udf'
                newtile['type']='tile_daoe_'+{'5up':'3up','3up':'5up'}[tile['type'][-3:]]
                newtile['ends']=[tile['ends'][x] for x in (3,2,1,0)]
                rotatedtiles.append(newtile)
                newtile = copy.copy(tile)
                newtile['name']+='_bf'
                newtile['type']='tile_daoe_'+{'5up':'3up','3up':'5up'}[tile['type'][-3:]]
                newtile['ends']=[tile['ends'][x] for x in (2,3,0,1)]
                rotatedtiles.append(newtile)
            elif tile['type'] == 'tile_daoe_doublehoriz_35up':
                newtile = copy.copy(tile)
                newtile['name']+='_lrf'
                newtile['type']='tile_daoe_doublevert_53up'
                newtile['ends']=[tile['ends'][x] for x in (2,1,0,5,4,3)]
                rotatedtiles.append(newtile)
                newtile = copy.copy(tile)
                newtile['name']+='_udf'
                newtile['type']='tile_daoe_doublevert_53up'
                newtile['ends']=[tile['ends'][x] for x in (5,4,3,2,1,0)]
                rotatedtiles.append(newtile)
                newtile = copy.copy(tile)
                newtile['name']+='_bf'
                newtile['ends']=[tile['ends'][x] for x in (3,4,5,0,1,2)]
                rotatedtiles.append(newtile)
            elif tile['type'] == 'tile_daoe_doublevert_35up':
                newtile = copy.copy(tile)
                newtile['name']+='_lrf'
                newtile['type']='tile_daoe_doublehoriz_53up'
                newtile['ends']=[tile['ends'][x] for x in (2,1,0,5,4,3)]
                rotatedtiles.append(newtile)
                newtile = copy.copy(tile)
                newtile['name']+='_udf'
                newtile['type']='tile_daoe_doublehoriz_53up'
                newtile['ends']=[tile['ends'][x] for x in (5,4,3,2,1,0)]
                rotatedtiles.append(newtile)
                newtile = copy.copy(tile)
                newtile['name']+='_bf'
                newtile['ends']=[tile['ends'][x] for x in (3,4,5,0,1,2)]
                rotatedtiles.append(newtile)

        ts['tiles'] += rotatedtiles
    
    for tile in ts['tiles']:
        if tile['type'] == 'tile_daoe_3up' or tile['type'] == 'tile_daoe_5up':
            newtile = {}
            newtile['edges'] = [ re.sub('/','_c',x) for x in tile['ends'] ]
            if 'name' in tile: newtile['name'] = tile['name']
            if 'conc' in tile: newtile['stoic'] = tile['conc']
            if 'color' in tile: newtile['color'] = tile['color']
            newtiles.append(newtile)

        if tile['type'] == 'tile_daoe_doublehoriz_35up' or tile['type'] == 'tile_daoe_doublehoriz_53up':
            newtile1 = {}
            newtile2 = {}
            newtile1['edges'] = [ re.sub('/','_c',x) for x in tile['ends'][0:1] ] \
                + [ tile['name']+'_db' ] \
                + [ re.sub('/','_c',x) for x in tile['ends'][4:] ]
            newtile2['edges'] = [ re.sub('/','_c',x) for x in tile['ends'][1:4] ] \
                + [ tile['name']+'_db' ]            
            newtile1['name'] = tile['name']+'_left'
            newtile2['name'] = tile['name']+'_right'
                        
            doubleends.append( tile['name']+'_db' )
            doubles.append( (newtile1['name'], newtile2['name']) )
            
            if 'conc' in tile: 
                newtile1['stoic'] = tile['conc']
                newtile2['stoic'] = tile['conc']
                
            if 'color' in tile: 
                newtile1['color'] = tile['color']
                newtile2['color'] = tile['color']
                
            newtiles.append(newtile1)
            newtiles.append(newtile2)
        if tile['type'] == 'tile_daoe_doublevert_35up' or tile['type'] == 'tile_daoe_doublevert_53up':
            newtile1 = {}
            newtile2 = {}
            newtile1['edges'] = [ re.sub('/','_c',x) for x in tile['ends'][0:2] ] \
                + [ tile['name']+'_db' ] \
                + [ re.sub('/','_c',x) for x in tile['ends'][5:] ]
            newtile2['edges'] = [ tile['name']+'_db' ] + [ re.sub('/','_c',x) for x in tile['ends'][2:5] ] 
            newtile1['name'] = tile['name']+'_top'
            newtile2['name'] = tile['name']+'_bottom'
                        
            vdoubleends.append( tile['name']+'_db' )
            vdoubles.append( (newtile1['name'], newtile2['name']) )
            
            if 'conc' in tile: 
                newtile1['stoic'] = tile['conc']
                newtile2['stoic'] = tile['conc']
                
            if 'color' in tile: 
                newtile1['color'] = tile['color']
                newtile2['color'] = tile['color']
                
            newtiles.append(newtile1)
            newtiles.append(newtile2)
    
    newends.append( { 'name': 'origami', 'strength': 100 } )
 
    for end in doubleends:
        newends.append( { 'name': end, 'strength': 10 } )
    for end in vdoubleends:
        newends.append( { 'name': end, 'strength': 10 } )

    gluelist = []
    if not perfect: 
        glueends = {'DT': [], 'TD': []}
        for end in ts['ends']:
            newends.append( { 'name': end['name'], 'strength': 0 } )
            newends.append( { 'name': end['name']+'_c', 'strength': 0 } )
            if (end['type'] == 'TD') or (end['type'] == 'DT'):
                glueends[end['type']].append((end['name'],end['fseq']))
                
        ef = en.energetics_santalucia(mismatchtype='max')
        
        for t in ['DT','TD']:
            names, fseqs = zip(*glueends[t])
            allnames = names + tuple( x+'_c' for x in names )
            ea = sd.endarray(fseqs, t)
            ar = sd.energy_array_uniform(ea,ef)
            for i1,n1 in enumerate(names):
                for i2,n2 in enumerate(allnames):
                    gluelist.append([n1,n2,float(ar[i1,i2])])
    else:
        if 'ends' not in ts.keys():
            ts['ends']=[]
        endsinlist = set( e['name'] for e in ts['ends'] )
        endsintiles = set()
        for tile in ts['tiles']:
            endsintiles.update( re.sub('/','',e) for e in tile['ends'] if e != 'hp')
        for end in ts['ends'] + list({'name': e} for e in endsintiles):
            newends.append( { 'name': end['name'], 'strength': 0 } )
            newends.append( { 'name': end['name']+'_c', 'strength': 0 } )
            gluelist.append([end['name'],end['name']+'_c',1.0]) 
            

    
    newends.append( {'name': 'hp', 'strength': 0} )

    xga = {}
    xga['doubletiles'] = [ list(x) for x in doubles ]
    xga['vdoubletiles'] = [ list(x) for x in vdoubles ]
    xga.update( ts['xgrow_options'] )
    xga.update( ts['xgrow_options'] )
     
        
    sts = { 'tiles': newtiles, 'bonds': newends, 'xgrowargs': xga, 'glues': gluelist }
    
    return sts

def from_yaml_tileadj( ts ):
    import stickydesign as sd
    import numpy as np

    # Combine ends and tile-specified adjacents
    newtiles = []
    newends = []
    endslist = set()
    doubleends = []
    doubles = []
    
    newtiles.append( { 'name': 'origami', 'edges': ['origami','origami','origami','origami'], 'stoic': 1e-9, 'color': 'white'} )
    
    for tile in ts['seed']['adapters']:
        newtile = {}
        newtile['edges'] = [ 'origami' ] +  [ re.sub('/','_c',x)+'_'+y for x,y in zip(tile['ends'],tile['adjs']) ] + [ 'origami' ]
        endslist.update( set(zip( newtile['edges'][1:3], tile['ends'], tile['adjs'] )) )
        if 'name' in tile: newtile['name'] = tile['name']
        newtile['stoic'] = 1e-9
        newtile['color'] = 'white'
        newtiles.append(newtile)    

    for tile in ts['tiles']:
        if tile['type'] == '3up' or tile['type'] == '5up':
            newtile = {}
            newtile['edges'] = [ re.sub('/','_c',x)+'_'+y for x,y in zip(tile['ends'],tile['adjs']) ]
            endslist.update( set(zip( newtile['edges'], tile['ends'], tile['adjs'] )) )
            if 'name' in tile: newtile['name'] = tile['name']
            if 'conc' in tile: newtile['stoic'] = tile['conc']
            if 'color' in tile: newtile['color'] = tile['color']
            newtiles.append(newtile)
        if tile['type'] == '3up5up' or tile['type'] == '5up3up':
            newtile1 = {}
            newtile2 = {}
            newtile1['edges'] = [ re.sub('/','_c',x)+'_'+y for x,y in zip(tile['ends'][0:1],tile['adjs'][0:1]) ] \
                + [ tile['name']+'_db' ] \
                + [ re.sub('/','_c',x)+'_'+y for x,y in zip(tile['ends'][4:],tile['adjs'][4:]) ]
            newtile2['edges'] = [ re.sub('/','_c',x)+'_'+y for x,y in zip(tile['ends'][1:4],tile['adjs'][1:4]) ] \
                + [ tile['name']+'_db' ]
            endslist.update( set(zip( [ re.sub('/','_c',x)+'_'+y for x,y in zip(tile['ends'],tile['adjs']) ], tile['ends'], tile['adjs'] )) )
            
            newtile1['name'] = tile['name']+'_left'
            newtile2['name'] = tile['name']+'_right'
                        
            doubleends.append( tile['name']+'_db' )
            doubles.append( (newtile1['name'], newtile2['name']) )
            
            if 'conc' in tile: 
                newtile1['stoic'] = tile['conc']
                newtile2['stoic'] = tile['conc']
                
            if 'color' in tile: 
                newtile1['color'] = tile['color']
                newtile2['color'] = tile['color']
                
            newtiles.append(newtile1)
            newtiles.append(newtile2)
    
    newends.append( { 'name': 'origami', 'strength': 100 } )
    
    for end in doubleends:
        newends.append( { 'name': end, 'strength': 100 } )
        
    for end in endslist:
        newends.append( { 'name': end[0], 'strength': 0 } )
    
    xga = {}
    xga['doubletiles'] = [ list(x) for x in doubles ]
    xga.update( ts['xgrow_options'] )
    
    
    # Now finally, deal with glues... in a perfect world first?
    # gluelist = []
    # for end1 in endslist:
    #     if end1[1][-1] == '/': continue
    #     for end2 in ( x for x in endslist if x[1]==end1[1]+'/' ):
    #         gluelist.append( [end1[0],end2[0],1] )
    #         
    ef = sd.energyfuncs_santalucia(mismatchtype='max')
    # 
    # gluelist = []
    # for end1 in endslist:
    #     if end1[1][-1] != '/':
    #         ed1 = [ x for x in ts['ends'] if x['name'] == end1[1] ][0]
    #     else:
    #         edd1 = [ x for x in ts['ends'] if x['name'] == end1[1][:-1] ][0]
    #         ed1 = { 'type': edd1['type'], 'seq': revcomp(edd1['seq'])}
    #     if (ed1['type'] == 'fake') or (ed1['type'] == 'fakeDT') or (ed1['type'] == 'fakeTD'):
    #         continue
    #     elif (ed1['type'] == 'DT'):
    #         e1 = sd.endarray([(end1[2]+ed1['seq']).lower()],'DT')
    #     elif (ed1['type'] == 'TD'):
    #         e1 = sd.endarray([(ed1['seq']+end1[2]).lower()],'TD')
    #     for end2 in endslist:
    #         if end2[1][-1] != '/':
    #             ed2 = [ x for x in ts['ends'] if x['name'] == end2[1] ][0]
    #         else:
    #             edd2 = [ x for x in ts['ends'] if x['name'] == end2[1][:-1] ][0]
    #             ed2 = { 'type': edd2['type'], 'seq': revcomp(edd2['seq'])}
    #         if ed1['type'] != ed2['type']:
    #             continue
    #         elif (ed1['type'] == 'DT'):
    #             e2 = sd.endarray([(end2[2]+ed2['seq']).lower()],'DT')
    #         elif (ed1['type'] == 'TD'):
    #             e2 = sd.endarray([(ed2['seq']+end2[2]).lower()],'TD')
    #         gluelist.append( [end1[0], end2[0], float(ef.uniform(e1,e2)[0])])
    dtl1 = []
    dtl2 = []
    tdl1 = []
    tdl2 = []
    dtn1 = []
    dtn2 = []
    tdn1 = []
    tdn2 = []
    for end1 in endslist:
        if end1[1][-1] != '/':
            ed1 = [ x for x in ts['ends'] if x['name'] == end1[1] ][0]
        else:
            edd1 = [ x for x in ts['ends'] if x['name'] == end1[1][:-1] ][0]
            ed1 = { 'type': edd1['type'], 'seq': revcomp(edd1['seq'])}
        if (ed1['type'] == 'fake') or (ed1['type'] == 'fakeDT') or (ed1['type'] == 'fakeTD'):
            continue
        for end2 in endslist:
            if end2[1][-1] != '/':
                ed2 = [ x for x in ts['ends'] if x['name'] == end2[1] ][0]
            else:
                edd2 = [ x for x in ts['ends'] if x['name'] == end2[1][:-1] ][0]
                ed2 = { 'type': edd2['type'], 'seq': revcomp(edd2['seq'])}
            if ed1['type'] != ed2['type']:
                continue
            elif (ed1['type'] == 'DT'):
                dtl1.append( (end1[2]+ed1['seq']).lower() )
                dtl2.append( (end2[2]+ed2['seq']).lower() )
                dtn1.append( end1[0] )
                dtn2.append( end2[0] )
            elif (ed1['type'] == 'TD'):
                tdl1.append( (ed1['seq']+end1[2]).lower() )
                tdl2.append( (ed2['seq']+end2[2]).lower() )
                tdn1.append( end1[0] )
                tdn2.append( end2[0] )
    
    dta1 = sd.endarray( dtl1, 'DT' )
    dta2 = sd.endarray( dtl2, 'DT' )
    tda1 = sd.endarray( tdl1, 'TD' )
    tda2 = sd.endarray( tdl2, 'TD' )
    
    dtg = ef.uniform(dta1, dta2)
    tdg = ef.uniform(tda1, tda2)
    
    dtgl = set([ tuple(sorted([x,y]) + [float(z)]) for x,y,z in zip(dtn1, dtn2, dtg) ])
    tdgl = set([ tuple(sorted([x,y]) + [float(z)]) for x,y,z in zip(tdn1, tdn2, tdg) ])
    
    gluelist = [ list(x) for x in dtgl.union(tdgl) ]
        
    sts = { 'tiles': newtiles, 'bonds': newends, 'xgrowargs': xga, 'glues': gluelist }
    
    return sts

def revcomp(seq):
    import stickydesign as sd
    import numpy as np
    seq = seq.lower()
    return "".join( reversed( [sd.wc[nt] for nt in seq] ) )
    



# vim: set sw=4 ts=4
import copy
import warnings
import re

import pkg_resources
import os

from . import tiletypes
from . import seeds
from . import util
from . import seq
from . import tilesets
from . import stickyends

from peppercompiler import compiler as compiler
from peppercompiler.design import spurious_design as spurious_design
from peppercompiler import finish as finish
from peppercompiler.DNA_classes import wc


def design_set(
        tileset,
        name='tsd_temp',
        includes=[pkg_resources.resource_filename(__name__, 'peppercomps-j1')],
        energetics=None,
        stickyopts={},
        reorderopts={},
        coreopts={},
        keeptemp=False):
    """Helper function to design sets from scratch, calling the numerous parts of
    tilesetdesigner. You may want to use the tilesetdesigner shell script
    instead.

    As with other functions in tilesetdesigner, this should not clobber inputs.

    :tileset: tileset definition dictionary, or an IO object with a read
    attribute, or a filename.

    :name: base name for temporary files (default tsd_temp)

    :returns: tileset definition dictionary, with added sequences

    """
    if energetics is None:
        energetics = stickyends.DEFAULT_ENERGETICS

    if hasattr(tileset, 'read'):
        tileset = tilesets.load_tileset_dict(tileset)
    else:
        tileset = tilesets.TileSet(tileset)

    tileset.check_consistent()
    tileset_with_ends_randomorder, new_ends = stickyends.create_sequences(
        tileset, energetics=energetics, **stickyopts)
    tileset_with_ends_ordered = stickyends.reorder(
        tileset_with_ends_randomorder,
        newends=new_ends,
        energetics=energetics,
        **reorderopts)
    tileset_with_strands = create_strand_sequences(
        tileset_with_ends_ordered, name, includes=includes, **coreopts)

    if 'guards' in tileset_with_strands.keys():
        tileset_with_strands = create_guard_strand_sequences(
            tileset_with_strands)

    # FIXME: this is temporary, until we have a better way of deciding.
    if 'createseqs' in tileset_with_strands['seed'].keys():
        tileset_with_strands = create_adapter_sequences(tileset_with_strands)

    if not keeptemp:
        os.remove(name + '.fix')
        os.remove(name + '.mfe')
        os.remove(name + '.pil')
        os.remove(name + '.save')
        os.remove(name + '.seqs')
        os.remove(name + '.sys')

    tileset_with_strands.check_consistent()
    return tileset_with_strands


def create_strand_sequences(tileset,
                            basename,
                            includes=None,
                            spurious_pars="verboten_weak=1.5",
                            *options):
    """Given a tileset dictionary with sticky ends sequences, create core sequences
for tiles.
    """

    newtileset = copy.deepcopy(tileset)

    create_pepper_input_files(newtileset, basename)

    compiler.compiler(
        basename, [],
        basename + '.pil',
        basename + '.save',
        fixed_file=basename + '.fix',
        includes=includes,
        synth=True)

    spurious_design.design(
        basename,
        infilename=basename + '.pil',
        outfilename=basename + '.mfe',
        verbose=True,
        struct_orient=True,
        tempname=basename + '-temp',
        extra_pars=spurious_pars,
        findmfe=False,
        cleanup=False)

    if 'info' not in newtileset.keys():
        newtileset['info'] = {}

    with open(basename + '-temp.sp') as f:
        a = f.read()
        cdi = {}
        cdi['basename'] = basename
        cdi['score_verboten'] = float(
            re.findall(r'score_verboten\s+score\s+=\s+([+-]?[\d.,]+)', a)[1])
        cdi['score_spurious'] = float(
            re.findall(r'score_spurious\s+score\s+=\s+([+-]?[\d.,]+)', a)[1])
        cdi['score_bonds'] = float(
            re.findall(r'score_bonds\s+score\s+=\s+([+-]?[\d.,]+)', a)[1])
        cdi['score'] = float(
            re.findall(r'weighted score\s+=\s+([+-]?[\d.,]+)', a)[1])
        cdi['spurious_output'] = re.search(r"(?<=FINAL\n\n)[\w\W]+weighted.*",
                                           a, re.MULTILINE).group(0)

    newtileset['info']['core'] = cdi

    finish.finish(
        basename + '.save',
        designname=basename + '.mfe',
        seqsname=basename + '.seqs',
        strandsname=None,
        run_kin=False,
        cleanup=False,
        trials=0,
        time=0,
        temp=27,
        conc=1,
        spurious=False,
        spurious_time=0)  # FIXME: shouldn't need so many options.

    tileset_with_strands = load_pepper_output_files(newtileset, basename)

    # Ensure:
    util.merge_endlists(tileset['ends'],
                        tiletypes.endlist_from_tilelist(
                            tileset_with_strands['tiles']))  # Ends still fit
    for tile in tileset_with_strands['tiles']:
        oldtile = tiletypes.gettile(tileset, tile['name'])
        if 'fullseqs' in oldtile.keys():
            for old, new in zip(oldtile['fullseqs'], tile['fullseqs']):
                seq.merge(old, new)  # old tile sequences remain
        assert oldtile['ends'] == tile['ends']

    # Check that old end sequences remain
    util.merge_endlists(tileset['ends'], tileset_with_strands['ends'])

    return tileset_with_strands


def create_pepper_input_files(tileset, basename):
    # Are we creating adapters in Pepper?
    if seeds.seedtypes[tileset['seed']['type']].needspepper:
        seedclass = seeds.seedtypes[tileset['seed']['type']]
        createadapts = True
    else:
        createadapts = False

    fixedfile = open(basename + ".fix", 'w')
    # We first need to create a fixed sequence list/file for pepper.
    # Add fixed sticky end and adjacent tile sequences.
    for end in tileset['ends']:
        if 'fseq' not in end.keys():
            continue
        seq = end['fseq'][1:-1]
        if end['type'] == 'TD':
            adj = end['fseq'][-1]
            cadj = end['fseq'][0]  # FIXME: WAS [1], OFF BY ONE!
        elif end['type'] == 'DT':
            adj = end['fseq'][0]
            cadj = end['fseq'][-1]  # FIXME: WAS [1], OFF BY ONE!
        else:
            print("warning! end {} not recognized".format(end['name']))
        fixedfile.write(
            "signal e_{0} = {1}\n".format(end['name'], seq.upper()))
        fixedfile.write(
            "signal a_{0} = {1}\n".format(end['name'], adj.upper()))
        fixedfile.write(
            "signal c_{0} = {1}\n".format(end['name'], cadj.upper()))
        # If we are creating adapter tiles in Pepper, add origami-determined
        # sequences
    if createadapts:
        for i, core in enumerate(seedclass.cores, 1):
            fixedfile.write("signal origamicore_{0} = {1}\n".format(i, core))

    # Now we'll create the system file in parts.
    importlist = set()
    compstring = ""

    for tile in tileset['tiles']:
        e = [[], []]
        for end in tile['ends']:
            if (end == 'hp'):
                continue
                # skip hairpins, etc that aren't designed by stickydesign
            e[0].append('e_' + end.replace('/', '*'))
            if end[-1] == '/':
                a = 'c_' + end[:-1] + '*'
            else:
                a = 'a_' + end
            e[1].append(a)
        s1 = " + ".join(e[0])
        s2 = " + ".join(e[1])
        tiletype = tile['type']
        if 'extra' in tile.keys():
            tiletype += '_' + tile['extra']
        compstring += "component {} = {}: {} -> {}\n".format(
            tile['name'], tiletype, s1, s2)
        importlist.add(tiletype)
        if 'fullseqs' in tile.keys():
            fixedfile.write("structure {}-tile = ".format(tile['name']) +
                            "+".join([seq.upper()
                                      for seq in tile['fullseqs']]) + "\n")

    if createadapts:
        importlist, compstring = seedclass.create_pepper_input_files(
            tileset['seed'], importlist, compstring)

    with open(basename + '.sys', 'w') as sysfile:
        sysfile.write("declare system {}: ->\n\n".format(basename))
        sysfile.write("import " + ", ".join(importlist) + "\n\n")
        sysfile.write(compstring)


def load_pepper_output_files(tileset, basename):
    import re

    # Are we creating adapters in Pepper?
    # if seeds.seedtypes[tileset['seed']['type']].needspepper:
    #     seedclass = seeds.seedtypes[tileset['seed']['type']]
    #     createadapts = True

    tset = copy.deepcopy(tileset)

    seqsstring = open(basename + '.seqs').read()

    # FIXME: we should do more than just get these sequences. We should also
    # check that our ends, complements, adjacents, etc are still correct. But
    # this is a pretty low priority.  UPDATE for 2017: WOW, LOOK AT ME BEING AN
    # IDIOT IN THE COMMENTS - CGE

    for tile in tset['tiles']:
        pepperstrands = re.compile('strand ' + tile['name'] +
                                   '-([^ ]+) = ([^\n]+)').findall(seqsstring)
        tile['fullseqs'] = tiletypes.order_pepper_strands(pepperstrands)

    for adapter in tset['seed']['adapters']:
        pepperstrands = re.compile('strand ' + adapter['name'] +
                                   '-([^ ]+) = ([^\n]+)').findall(seqsstring)
        adapter['fullseqs'] = tiletypes.order_pepper_strands(pepperstrands)

    return tset


def create_guard_strand_sequences(tileset):
    tset = copy.deepcopy(tileset)

    for guard in tset['guards']:
        tile = tiletypes.gettile(tset, guard[0])
        guard.append(wc(tile['fullseqs'][guard[1] - 1]))

    return tset


def create_adapter_sequence_diagrams(tileset, filename, *options):
    from lxml import etree
    import pkg_resources
    import os.path

    base = etree.parse(
        pkg_resources.resource_stream(
            __name__, os.path.join('seqdiagrambases', 'blank.svg')))
    baseroot = base.getroot()
    pos = 200
    for adapterdef in tileset['seed']['adapters']:

        seedclass = seeds.seedtypes[tileset['seed']['type']]
        group = seedclass.create_adapter_sequence_diagram(adapterdef)

        group.attrib['transform'] = 'translate(0,{})'.format(pos)
        pos += 200
        baseroot.append(group)

    base.write(filename)


def create_sequence_diagrams(tileset, filename, *options):
    from lxml import etree
    import pkg_resources
    import os.path

    base = etree.parse(
        pkg_resources.resource_stream(
            __name__, os.path.join('seqdiagrambases', 'blank.svg')))
    baseroot = base.getroot()
    pos = 150
    for tiledef in tileset['tiles']:

        tile = tiletypes.tfactory.parse(tiledef)
        group = tile.sequence_diagram()

        group.attrib['transform'] = 'translate(0,{})'.format(pos)
        pos += 150
        baseroot.append(group)

    base.write(filename)


def create_adapter_sequences(tileset):
    seedclass = seeds.seedtypes[tileset['seed']['type']]
    if seedclass.needspepper:
        warnings.warn(
            "This set must have adapter sequences created during\
 regular sequence design. You can ignore this if you just created sequences."
        )
        return tileset
    return seedclass.create_adapter_sequences(tileset)


def create_abstract_diagrams(tileset, filename, *options):
    import svgwrite

    drawing = svgwrite.Drawing(filename)

    pos = 0
    lim = 10000

    for tiledef in tileset['tiles']:
        tile = tiletypes.tfactory.parse(tiledef)

        group = tile.abstract_diagram(drawing)

        group['transform'] = "translate({},{})".format((pos % lim) * 150,
                                                       pos / lim * 150)
        pos += 1

        drawing.add(group)
        pos += 1

    drawing.save()


def create_strand_order_list(tileset):
    tf = [tiletypes.tfactory.parse(t) for t in tileset['tiles']]
    return sum((x.orderableseqs for x in tf), [])


def create_layout_diagrams(tileset, xgrowarray, filename, scale=1, *options):
    import svgwrite
    from . import stxg

    b = svgwrite.Drawing(filename)

    c = b.g()

    svgtiles = {}

    for tiledef in tileset['tiles']:
        tile = tiletypes.tfactory.parse(tiledef)

        group = tile.abstract_diagram(b)
        svgtiles[tile['name']] = group

    tilelist = stxg.from_yaml_endadj(tileset)['tiles']
    tilen = [None] + [x['name'] for x in tilelist]
    firstxi = 10000
    firstyi = 10000

    for yi in range(0, xgrowarray.shape[0]):
        for xi in range(0, xgrowarray.shape[1]):
            tn = tilen[xgrowarray[yi, xi]]
            if tn and tn[-5:] == '_left':
                tn = tn[:-5]
            if tn and tn[-4:] == '_top':
                tn = tn[:-4]
            if not (tn in svgtiles.keys()):
                continue
            if xi < firstxi:
                firstxi = xi
            if yi < firstyi:
                firstyi = yi
            st = svgtiles[tn].copy()
            st['transform'] = 'translate({},{})'.format(xi * 100, yi * 100)
            c.add(st)

    c['transform'] = "translate({},{}) scale({})".format(
        -firstxi * 100 * scale, -firstyi * 100 * scale, scale)

    b.add(c)
    b.save()



from .tilesets import TileSet
from . import tiletypes
from . import util

import stickydesign as sd
import numpy as np
import collections
from random import shuffle
from datetime import datetime, timezone

import logging

DEFAULT_ENERGETICS = sd.EnergeticsDAOE(
    temperature=33, mismatchtype='combined', coaxparams=True)

DEFAULT_MULTIMODEL_ENERGETICS = [
    sd.EnergeticsDAOE(
        temperature=33, mismatchtype='combined', coaxparams='protozanova'),
    sd.EnergeticsDAOE(
        temperature=33, mismatchtype='combined', coaxparams='pyshni'),
    sd.EnergeticsDAOE(
        temperature=33, mismatchtype='combined', coaxparams='peyret'),
    sd.EnergeticsDAOE(
        temperature=33, mismatchtype='combined', coaxparams=False)]

DEFAULT_MM_ENERGETICS_NAMES = ['Prot', 'Pysh', 'Peyr', 'None']

SELOGGER = logging.getLogger(__name__)


def create_sequences(tileset, method='default', energetics=None,
                     trials=100, sdopts={}, ecpars={}):
    """Create sticky end sequences for a tileset, using stickydesign.  This new
version should be more flexible, and should be able to keep old sticky ends,
accounting for them in creating new ones.

Parameters
----------

tileset: the tileset to create sticky ends sequences for.  This will be copied
     and returned, not modified.

method: [default 'default'] if 'default', use the default, single-model
sequence design.  If 'multimodel', use multimodel end choice.

energetics: the energetics instance to use for the design.  If None (default),
will use alhambra.designer.DEFAULT_ENERGETICS.  Note that if
method='multimodel', energetics *must* be specified, and should be a list of
energy models.  The first model in the list will be used as primary.

Outputs (tileset, new_ends) where new_ends is a list of new end names that
were designed.
"""
    info = {}
    info['method'] = method
    info['time'] = datetime.now(tz=timezone.utc).isoformat()
    info['sd_version'] = sd.version.__version__

    if not energetics:
        if method == 'multimodel':
            energetics = DEFAULT_MULTIMODEL_ENERGETICS
        else:
            energetics = DEFAULT_ENERGETICS
    if method == 'multimodel' and not isinstance(energetics,
                                                 collections.Iterable):
        raise ValueError("Energetics must be an iterable for multimodel.")
    elif method == 'multimodel':
        all_energetics = energetics
        energetics = all_energetics[0]
        info['energetics'] = [str(e) for e in all_energetics]
        info['trails'] = trials
    elif method == 'default':
        info['energetics'] = str(energetics)

    # Steps for doing this:

    # Create a copy of the tileset.
    newtileset = TileSet(tileset).copy()

    # Build a list of ends from the endlist in the tileset.  Do this
    # by creating a named_list, then merging them into it.
    ends = util.named_list()

    if 'ends' in newtileset.keys():
        ends = util.merge_endlists(
            ends, newtileset['ends'], fail_immediate=False, in_place=True)

    # This is the endlist from the tiles themselves.
    if 'tiles' in newtileset.keys():  # maybe you just want ends?
        # this checks for end/complement usage, and whether any
        # previously-describedends are unused
        # FIXME: implement
        # tiletypes.check_end_usage(newtileset['tiles'], ends)

        endlist_from_tiles = tiletypes.endlist_from_tilelist(
            newtileset['tiles'])

    ends = util.merge_endlists(ends, endlist_from_tiles, in_place=True)

    # Ensure that if there are any resulting completely-undefined ends, they
    # have their sequences removed.
    for end in ends:
        if 'fseq' in end.keys() and end['fseq'] == 'nnnnnnn':
            del (end['fseq'])

    # Build inputs suitable for stickydesign: lists of old sequences for TD/DT,
    # and numbers of new sequences needed.
    oldDTseqs = [
        end['fseq'] for end in ends
        if 'fseq' in end.keys() and end['type'] == 'DT'
    ]
    oldTDseqs = [
        end['fseq'] for end in ends
        if 'fseq' in end.keys() and end['type'] == 'TD'
    ]

    newTDnames = [
        end['name'] for end in ends
        if 'fseq' not in end.keys() and end['type'] == 'TD'
    ]
    newDTnames = [
        end['name'] for end in ends
        if 'fseq' not in end.keys() and end['type'] == 'DT'
    ]

    # Deal with energetics, considering potential old sequences.
    # FIXME: EXPLAIN WHAT THIS ABSTRUSE CODE DOES...
    # TODO: tests needs to test this
    targets = []
    if len(oldDTseqs) == 0 and len(oldTDseqs) == 0:
        targets.append(sd.enhist('DT', 5, energetics=energetics)[2]['emedian'])
        targets.append(sd.enhist('TD', 5, energetics=energetics)[2]['emedian'])
    if len(oldDTseqs) > 0:
        targets.append(
            energetics.matching_uniform(sd.endarray(oldDTseqs, 'DT')))
    if len(oldTDseqs) > 0:
        targets.append(
            energetics.matching_uniform(sd.endarray(oldTDseqs, 'TD')))
    targetint = np.average(targets)

    if method == 'multimodel' and (len(oldDTseqs) > 0 or len(oldTDseqs) > 0):
        raise NotImplementedError()

    if method == 'default':
        # Create new sequences.
        newTDseqs = sd.easyends(
            'TD',
            5,
            number=len(newTDnames),
            energetics=energetics,
            interaction=targetint,
            **sdopts).tolist()

        newDTseqs = sd.easyends(
            'DT',
            5,
            number=len(newDTnames),
            energetics=energetics,
            interaction=targetint,
            **sdopts).tolist()

    elif method == 'multimodel':
        SELOGGER.info(
            "starting multimodel sticky end generation " +
            "of TD ends for {} DT and {} TD ends, {} trials.".format(
                len(newDTnames), len(newTDnames), trials))

        endchooser = sd.multimodel.endchooser(all_energetics, **ecpars)

        newTDseqs = []
        pl = util.ProgressLogger(SELOGGER, trials*2)
        for i in range(0, trials):
            newTDseqs.append(
                sd.easyends(
                    'TD',
                    5,
                    number=len(newTDnames),
                    energetics=energetics,
                    interaction=targetint,
                    echoose=endchooser,
                    **sdopts))
            pl.update(i)

        tvals = [[e.matching_uniform(x[0:1])
                 for e in all_energetics] for x in newTDseqs]
        endchoosers = [sd.multimodel.endchooser(all_energetics,
                                                target_vals=tval,
                                                **ecpars)
                       for tval in tvals]

        SELOGGER.info("generating corresponding DT ends")
        newDTseqs = []
        for i, echoose in enumerate(endchoosers):
            newDTseqs.append(
                sd.easyends(
                    'DT',
                    5,
                    number=len(newDTnames),
                    energetics=energetics,
                    interaction=targetint,
                    echoose=echoose,
                    **sdopts))
            pl.update(i+trials)

        scores = [sd.multimodel.deviation_score(list(e), all_energetics)
                  for e in zip(newTDseqs, newDTseqs)]

        sort = np.argsort(scores)

        newTDseqs = newTDseqs[sort[0]].tolist()
        newDTseqs = newDTseqs[sort[0]].tolist()
        info['score'] = scores[sort[0]]
        info['maxscore'] = scores[sort[-1]]
        info['meanscore'] = np.mean(scores)

    # FIXME: move to stickydesign
    assert len(newTDseqs) == len(newTDnames)
    assert len(newDTseqs) == len(newDTnames)

    # Shuffle the lists of end sequences, to ensure that they're random order,
    # and that ends used earlier in the set are not always better than those
    # used later.
    shuffle(newTDseqs)
    shuffle(newDTseqs)

    for name, s in zip(newDTnames, newDTseqs):
        ends[name]['fseq'] = s
    for name, s in zip(newTDnames, newTDseqs):
        ends[name]['fseq'] = s

    ends.check_consistent()

    # Ensure that the old and new sets have consistent end definitions,
    # and that the tile definitions still fit.
    tiletypes.merge_endlists(tileset['ends'], ends)
    tiletypes.merge_endlists(
        tiletypes.endlist_from_tilelist(newtileset['tiles']), ends)

    # Apply new sequences to tile system.
    newtileset['ends'] = ends
    if 'info' not in newtileset.keys():
        newtileset['info'] = {}
    newtileset['info']['end_design'] = info

    return (newtileset, newTDnames + newDTnames)


def reorder(tileset,
            newends=[],
            hightemp=0.1,
            lowtemp=1e-7,
            steps=45000,
            update=1000,
            energetics=None):
    """Given a tileset dictionary that includes sticky end sequences, reorder these
    to try to optimize error rates.
    """
    # from . import endreorder
    from . import endreorder_fast
    from . import anneal

    if energetics is None:
        energetics = DEFAULT_ENERGETICS

    tset = TileSet(tileset).copy()

    if 'info' not in tset.keys():
        tset['info'] = {}

    reordersys = endreorder_fast.EndSystemFseq(
        tset, newends, energetics=energetics)

    # reordersys_old = endreorder.EndSystemFseq( tset, energetics=energetics )

    # FIXME: better parameter control here.
    annealer = anneal.Annealer(reordersys.score, reordersys.mutate)

    newstate = annealer.anneal(reordersys.initstate, hightemp, lowtemp, steps,
                               update)

    # Now take that new state, and apply it to the new tileset.
    seqs = reordersys.slowseqs(newstate[0])
    for end in tset['ends']:
        if end['type'] in ['DT', 'TD']:
            eloc = reordersys.enlocs[end['name']]
            end['fseq'] = seqs[eloc[1]].tolist()[eloc[0]]

    ri = {}

    ri['score'] = reordersys.score(newstate[0])

    tset['info']['reorder'] = ri

    # Ensure that only ends in newends moved: that all others remain mergeable:
    if newends:
        old_ends_from_new_set = util.named_list(end for end in tset['ends']
                                                if end['name'] not in newends)
        util.merge_endlists(tileset['ends'], old_ends_from_new_set)

    # Ensure system consistency
    tset.check_consistent()
    return tset
